"""Fixtures for Regionalizers."""

import geopandas as gpd
import pytest
from shapely import geometry

from srai.constants import GEOMETRY_COLUMN, WGS84_CRS


@pytest.fixture  # type: ignore
def gdf_empty() -> gpd.GeoDataFrame:
    """Get empty GeoDataFrame."""
    return gpd.GeoDataFrame()


@pytest.fixture  # type: ignore
def gdf_no_crs() -> gpd.GeoDataFrame:
    """Get GeoDataFrame with no crs."""
    return gpd.GeoDataFrame(
        geometry=[
            geometry.Polygon(
                shell=[
                    (-1, 0),
                    (0, 0.5),
                    (1, 0),
                    (1, 1),
                    (0, 1),
                ]
            )
        ]
    )


@pytest.fixture  # type: ignore
def gdf_polygons() -> gpd.GeoDataFrame:
    """Get GeoDataFrame with two polygons."""
    return gpd.GeoDataFrame(
        geometry=[
            geometry.Polygon(
                shell=[
                    (-1, 0),
                    (0, 0.5),
                    (1, 0),
                    (1, 1),
                    (0, 1),
                ],
                holes=[
                    [
                        (0.8, 0.9),
                        (0.9, 0.55),
                        (0.8, 0.3),
                        (0.5, 0.4),
                    ]
                ],
            ),
            geometry.Polygon(shell=[(-0.25, 0), (0.25, 0), (0, 0.2)]),
        ],
        crs=WGS84_CRS,
    )


@pytest.fixture  # type: ignore
def gdf_multipolygon() -> gpd.GeoDataFrame:
    """Get GeoDataFrame with multipolygon."""
    return gpd.GeoDataFrame(
        geometry=[
            geometry.MultiPolygon(
                [
                    (
                        [
                            (-1, 0),
                            (0, 0.5),
                            (1, 0),
                            (1, 1),
                            (0, 1),
                        ],
                        (
                            [
                                [
                                    (0.8, 0.9),
                                    (0.9, 0.55),
                                    (0.8, 0.3),
                                    (0.5, 0.4),
                                ]
                            ]
                        ),
                    ),
                    (
                        [(-0.25, 0), (0.25, 0), (0, 0.2)],
                        (),
                    ),
                ]
            )
        ],
        crs=WGS84_CRS,
    )


@pytest.fixture  # type: ignore
def earth_poles() -> list[geometry.Point]:
    """Get 6 Earth poles."""
    return [
        geometry.Point(0, 0),
        geometry.Point(90, 0),
        geometry.Point(180, 0),
        geometry.Point(-90, 0),
        geometry.Point(0, 90),
        geometry.Point(0, -90),
    ]


@pytest.fixture  # type: ignore
def gdf_earth_poles(earth_poles) -> gpd.GeoDataFrame:
    """Get GeoDataFrame with 6 Earth poles."""
    return gpd.GeoDataFrame(
        {GEOMETRY_COLUMN: earth_poles},
        index=[1, 2, 3, 4, 5, 6],
        crs=WGS84_CRS,
    )


@pytest.fixture  # type: ignore
def gdf_poland() -> gpd.GeoDataFrame:
    """Get Poland GeoDataFrame."""
    world = gpd.read_file(gpd.datasets.get_path("naturalearth_lowres"))
    area = world[world.name == "Poland"]
    return area


@pytest.fixture  # type: ignore
def earth_bbox() -> geometry.Polygon:
    """Get full bounding box GeoDataFrame."""
    return geometry.box(minx=-180, maxx=180, miny=-90, maxy=90)


@pytest.fixture  # type: ignore
def gdf_earth_bbox(earth_bbox) -> gpd.GeoDataFrame:
    """Get full bounding box GeoDataFrame."""
    return gpd.GeoDataFrame({GEOMETRY_COLUMN: [earth_bbox]}, crs=WGS84_CRS)



"""
Base OSM groups filter.

This module contains the grouped OSM tags filter that was used in ARIC@SIGSPATIAL 2021 paper [1].

References:
    1. https://doi.org/10.1145/3486626.3493434
    1. https://arxiv.org/abs/2111.00990
"""

from srai.loaders.osm_loaders.filters import GroupedOsmTagsFilter

BASE_OSM_GROUPS_FILTER: GroupedOsmTagsFilter = {
    "water": {"natual": ["water", "bay", "beach", "coastline"], "waterway": ["riverbank"]},
    "aerialway": {
        "aerialway": [
            "cable_car",
            "gondola",
            "mixed_lift",
            "chair_lift",
            "drag_lift",
            "t-bar",
            "j-bar",
            "platter",
            "rope_tow",
            "magic_carpet",
            "zip_line",
            "goods",
            "station",
        ]
    },
    "airports": {"aeroway": ["aerodrome", "heliport", "spaceport"]},
    "sustenance": {
        "amenity": [
            "bar",
            "bbq",
            "biergarten",
            "cafe",
            "fast_food",
            "food_court",
            "ice_cream",
            "pub",
            "restaurant",
        ]
    },
    "education": {
        "amenity": [
            "college",
            "driving_school",
            "kindergarten",
            "language_school",
            "library",
            "toy_library",
            "music_school",
            "school",
            "university",
        ]
    },
    "transportation": {
        "amenity": [
            "bicycle_parking",
            "bicycle_repair_station",
            "bicycle_rental",
            "boat_rental",
            "boat_sharing",
            "car_rental",
            "car_sharing",
            "car_wash",
            "charging_station",
            "bus_stop",
            "ferry_terminal",
            "fuel",
            "motorcycle_parking",
            "parking",
            "taxi",
            "bus_station",
        ],
        "public_transport": ["station", "stop_position"],
        "railway": ["station", "subway_entrance", "tram_stop"],
        "building": ["train_station"],
        "highway": ["bus_stop"],
    },
    "finances": {"amenity": ["atm", "bank", "bureau_de_change"]},
    "healthcare": {
        "amenity": [
            "baby_hatch",
            "clinic",
            "dentist",
            "doctors",
            "hospital",
            "nursing_home",
            "pharmacy",
            "social_facility",
            "veterinary",
        ]
    },
    "culture_art_entertainment": {
        "amenity": [
            "arts_centre",
            "brothel",
            "casino",
            "cinema",
            "community_centre",
            "gambling",
            "nightclub",
            "planetarium",
            "public_bookcase",
            "social_centre",
            "stripclub",
            "studio",
            "theatre",
        ]
    },
    "other": {
        "amenity": [
            "animal_boarding",
            "animal_shelter",
            "childcare",
            "conference_centre",
            "courthouse",
            "crematorium",
            "embassy",
            "fire_station",
            "grave_yard",
            "internet_cafe",
            "marketplace",
            "monastery",
            "place_of_worship",
            "police",
            "post_office",
            "prison",
            "ranger_station",
            "refugee_site",
            "townhall",
        ]
    },
    "buildings": {
        "building": ["commercial", "industrial", "warehouse"],
        "office": True,
        "waterway": ["dock", "boatyard"],
    },
    "emergency": {"emergency": ["ambulance_station", "defibrillator", "landing_site"]},
    "historic": {
        "historic": [
            "aqueduct",
            "battlefield",
            "building",
            "castle",
            "church",
            "citywalls",
            "fort",
            "memorial",
            "monastery",
            "monument",
            "ruins",
            "tower",
        ]
    },
    "leisure": {
        "leisure": [
            "adult_gaming_centre",
            "amusement_arcade",
            "beach_resort",
            "common",
            "dance",
            "dog_park",
            "escape_game",
            "fitness_centre",
            "fitness_station",
            "garden",
            "hackerspace",
            "horse_riding",
            "ice_rink",
            "marina",
            "miniature_golf",
            "nature_reserve",
            "park",
            "pitch",
            "slipway",
            "sports_centre",
            "stadium",
            "summer_camp",
            "swimming_area",
            "swimming_pool",
            "track",
            "water_park",
        ],
        "amenity": ["public_bath", "dive_centre"],
    },
    "shops": {"shop": True},
    "sport": {"sport": True},
    "tourism": {"tourism": True},
    "greenery": {
        "leisure": ["park"],
        "natural": ["grassland", "scrub"],
        "landuse": [
            "grass",
            "allotments",
            "forest",
            "flowerbed",
            "meadow",
            "village_green",
            "grassland",
            "scrub",
            "garden",
            "park",
            "recreation_ground",
        ],
    },
}



"""Tests for LookupNeighbourhood."""

from typing import Any, Optional, TypeVar

import pytest

from srai.neighbourhoods import Neighbourhood

T = TypeVar("T")


class LookupNeighbourhood(Neighbourhood[T]):
    """LookupNeighbourhood."""

    def __init__(self, lookup: dict[T, set[T]], include_center: bool = False) -> None:
        """
        LookupNeighbourhood constructor.

        Args:
            lookup (Dict[T, Set[T]]): Mapping of region to its neighbours.
            include_center (bool): Whether to include the region itself in the neighbours.
        """
        super().__init__(include_center)
        self.lookup = lookup

    def get_neighbours(self, index: T, include_center: Optional[bool] = None) -> set[T]:
        """
        Get neighbours for region at index.

        Args:
            index (T): Index of region in mapping.
            include_center (Optional[bool]): Whether to include the region itself in the neighbours.
            If None, the value set in __init__ is used. Defaults to None.
        """
        neighbours = self.lookup[index]
        return self._handle_center(
            index, 1, neighbours, at_distance=False, include_center_override=include_center
        )


@pytest.fixture  # type: ignore
def grid_3_by_3_neighbourhood() -> dict[int, set[int]]:
    """
    Get grid neighbourhood.

    This dict represents a simple 3 by 3 grid-like neighbourhood. The tiles are numbered from 1 to
    9, from left to right, top to bottom. The tiles are considered neighbours if they are adjacent
    by edge, not by vertex. Visually it looks like this:

    [[1, 2, 3],
     [4, 5, 6],
     [7, 8, 9]]

    Returns:
        Dict[int, Set[int]]: A dict representing 3 by 3 grid neighbourhood.
    """
    return {
        1: {2, 4},
        2: {1, 3, 5},
        3: {2, 6},
        4: {1, 5, 7},
        5: {2, 4, 6, 8},
        6: {3, 5, 9},
        7: {4, 8},
        8: {5, 7, 9},
        9: {6, 8},
    }


@pytest.fixture  # type: ignore
def grid_3_by_3_irrregular_neighbourhood() -> dict[int, set[int]]:
    """
    Get irregular grid neighbourhood.

    This dict represents a simple 3 by 3 grid-like neighbourhood. The tiles are numbered from 1 to
    4, from left to right, top to bottom. The tiles are considered neighbours if they are adjacent
    by edge, not by vertex. Tiles are irregular, not single squares 1 by 1. Visually it looks like
    this:

    [[1, 1, 2],
     [1, 1, 2],
     [3, 4, 4]]

    Returns:
        Dict[int, Set[int]]: A dict representing 3 by 3 grid neighbourhood.
    """
    return {
        1: {2, 3, 4},
        2: {1, 4},
        3: {1, 4},
        4: {1, 2, 3},
    }


@pytest.mark.parametrize(  # type: ignore
    "neighbourhood_fixture,index,expected",
    [
        ("grid_3_by_3_neighbourhood", 1, {2, 4}),
        ("grid_3_by_3_neighbourhood", 2, {1, 3, 5}),
        ("grid_3_by_3_neighbourhood", 3, {2, 6}),
        ("grid_3_by_3_neighbourhood", 4, {1, 5, 7}),
        ("grid_3_by_3_neighbourhood", 5, {2, 4, 6, 8}),
        ("grid_3_by_3_neighbourhood", 6, {3, 5, 9}),
        ("grid_3_by_3_neighbourhood", 7, {4, 8}),
        ("grid_3_by_3_neighbourhood", 8, {5, 7, 9}),
        ("grid_3_by_3_neighbourhood", 9, {6, 8}),
        ("grid_3_by_3_irrregular_neighbourhood", 1, {2, 3, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 2, {1, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 3, {1, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 4, {1, 2, 3}),
    ],
)
def test_get_neighbours(
    neighbourhood_fixture: str,
    index: str,
    expected: set[str],
    request: Any,
) -> None:
    """Test get_neighbours with overriding include_center."""
    # Test with class include_center=False
    neighbourhood_data = request.getfixturevalue(neighbourhood_fixture)
    expected_with_include_center = expected | {index}

    neighbourhood = LookupNeighbourhood(neighbourhood_data)
    neighbours = neighbourhood.get_neighbours(index)
    assert neighbours == expected

    neighbours = neighbourhood.get_neighbours(index, include_center=True)
    assert neighbours == expected_with_include_center

    # Test with class include_center=True
    neighbourhood = LookupNeighbourhood(neighbourhood_data, include_center=True)
    neighbours = neighbourhood.get_neighbours(index)
    assert neighbours == expected_with_include_center

    neighbours = neighbourhood.get_neighbours(index, include_center=False)
    assert neighbours == expected


@pytest.mark.parametrize(  # type: ignore
    "neighbourhood_fixture, index, distance, expected, expected_with_include_center",
    [
        ("grid_3_by_3_neighbourhood", 5, -2, set(), set()),
        ("grid_3_by_3_neighbourhood", 5, -1, set(), set()),
        ("grid_3_by_3_neighbourhood", 5, 0, set(), {5}),
        ("grid_3_by_3_neighbourhood", 5, 1, {2, 4, 6, 8}, {2, 4, 6, 8}),
        ("grid_3_by_3_neighbourhood", 5, 2, {1, 3, 7, 9}, {1, 3, 7, 9}),
        ("grid_3_by_3_neighbourhood", 5, 3, set(), set()),
        ("grid_3_by_3_neighbourhood", 4, 0, set(), {4}),
        ("grid_3_by_3_neighbourhood", 4, 1, {1, 5, 7}, {1, 5, 7}),
        ("grid_3_by_3_neighbourhood", 4, 2, {2, 6, 8}, {2, 6, 8}),
        ("grid_3_by_3_neighbourhood", 4, 3, {3, 9}, {3, 9}),
        ("grid_3_by_3_neighbourhood", 4, 4, set(), set()),
        ("grid_3_by_3_neighbourhood", 1, 0, set(), {1}),
        ("grid_3_by_3_neighbourhood", 1, 1, {2, 4}, {2, 4}),
        ("grid_3_by_3_neighbourhood", 1, 2, {3, 5, 7}, {3, 5, 7}),
        ("grid_3_by_3_neighbourhood", 1, 3, {6, 8}, {6, 8}),
        ("grid_3_by_3_neighbourhood", 1, 4, {9}, {9}),
        ("grid_3_by_3_neighbourhood", 1, 5, set(), set()),
        ("grid_3_by_3_irrregular_neighbourhood", 1, -2, set(), set()),
        ("grid_3_by_3_irrregular_neighbourhood", 1, -1, set(), set()),
        ("grid_3_by_3_irrregular_neighbourhood", 1, 0, set(), {1}),
        ("grid_3_by_3_irrregular_neighbourhood", 1, 1, {2, 3, 4}, {2, 3, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 1, 2, set(), set()),
        ("grid_3_by_3_irrregular_neighbourhood", 2, 0, set(), {2}),
        ("grid_3_by_3_irrregular_neighbourhood", 2, 1, {1, 4}, {1, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 2, 2, {3}, {3}),
        ("grid_3_by_3_irrregular_neighbourhood", 2, 3, set(), set()),
        ("grid_3_by_3_irrregular_neighbourhood", 3, 0, set(), {3}),
        ("grid_3_by_3_irrregular_neighbourhood", 3, 1, {1, 4}, {1, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 3, 2, {2}, {2}),
        ("grid_3_by_3_irrregular_neighbourhood", 3, 3, set(), set()),
        ("grid_3_by_3_irrregular_neighbourhood", 4, 0, set(), {4}),
        ("grid_3_by_3_irrregular_neighbourhood", 4, 1, {1, 2, 3}, {1, 2, 3}),
        ("grid_3_by_3_irrregular_neighbourhood", 4, 2, set(), set()),
    ],
)
def test_get_neighbours_at_distance(
    neighbourhood_fixture: str,
    index: str,
    distance: int,
    expected: set[str],
    expected_with_include_center: set[str],
    request: Any,
) -> None:
    """Test neighbours at distance."""
    neighbourhood_data = request.getfixturevalue(neighbourhood_fixture)
    neighbourhood = LookupNeighbourhood(neighbourhood_data)
    neighbours = neighbourhood.get_neighbours_at_distance(index, distance)
    assert neighbours == expected

    neighbours = neighbourhood.get_neighbours_at_distance(index, distance, include_center=True)
    assert neighbours == expected_with_include_center

    neighbourhood = LookupNeighbourhood(neighbourhood_data, include_center=True)
    neighbours = neighbourhood.get_neighbours_at_distance(index, distance)
    assert neighbours == expected_with_include_center

    neighbours = neighbourhood.get_neighbours_at_distance(index, distance, include_center=False)
    assert neighbours == expected


@pytest.mark.parametrize(  # type: ignore
    "neighbourhood_fixture, index, distance, expected",
    [
        ("grid_3_by_3_neighbourhood", 5, -2, set()),
        ("grid_3_by_3_neighbourhood", 5, -1, set()),
        ("grid_3_by_3_neighbourhood", 5, 0, set()),
        ("grid_3_by_3_neighbourhood", 5, 1, {2, 4, 6, 8}),
        ("grid_3_by_3_neighbourhood", 5, 2, {1, 2, 3, 4, 6, 7, 8, 9}),
        ("grid_3_by_3_neighbourhood", 5, 3, {1, 2, 3, 4, 6, 7, 8, 9}),
        ("grid_3_by_3_neighbourhood", 4, 0, set()),
        ("grid_3_by_3_neighbourhood", 4, 1, {1, 5, 7}),
        ("grid_3_by_3_neighbourhood", 4, 2, {1, 2, 5, 6, 7, 8}),
        ("grid_3_by_3_neighbourhood", 4, 3, {1, 2, 3, 5, 6, 7, 8, 9}),
        ("grid_3_by_3_neighbourhood", 4, 4, {1, 2, 3, 5, 6, 7, 8, 9}),
        ("grid_3_by_3_neighbourhood", 1, 0, set()),
        ("grid_3_by_3_neighbourhood", 1, 1, {2, 4}),
        ("grid_3_by_3_neighbourhood", 1, 2, {2, 3, 4, 5, 7}),
        ("grid_3_by_3_neighbourhood", 1, 3, {2, 3, 4, 5, 6, 7, 8}),
        ("grid_3_by_3_neighbourhood", 1, 4, {2, 3, 4, 5, 6, 7, 8, 9}),
        ("grid_3_by_3_neighbourhood", 1, 5, {2, 3, 4, 5, 6, 7, 8, 9}),
        ("grid_3_by_3_irrregular_neighbourhood", 1, -2, set()),
        ("grid_3_by_3_irrregular_neighbourhood", 1, -1, set()),
        ("grid_3_by_3_irrregular_neighbourhood", 1, 0, set()),
        ("grid_3_by_3_irrregular_neighbourhood", 1, 1, {2, 3, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 1, 2, {2, 3, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 2, 0, set()),
        ("grid_3_by_3_irrregular_neighbourhood", 2, 1, {1, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 2, 2, {1, 3, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 2, 3, {1, 3, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 3, 0, set()),
        ("grid_3_by_3_irrregular_neighbourhood", 3, 1, {1, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 3, 2, {1, 2, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 3, 3, {1, 2, 4}),
        ("grid_3_by_3_irrregular_neighbourhood", 4, 0, set()),
        ("grid_3_by_3_irrregular_neighbourhood", 4, 1, {1, 2, 3}),
        ("grid_3_by_3_irrregular_neighbourhood", 4, 2, {1, 2, 3}),
    ],
)
def test_get_neighbours_up_to_distance(
    neighbourhood_fixture: str, index: str, distance: int, expected: set[str], request: Any
) -> None:
    """Test neighbours up to a distance."""
    neighbourhood_data = request.getfixturevalue(neighbourhood_fixture)
    expected_with_include_center = expected.copy()
    if distance >= 0:
        expected_with_include_center.add(index)
    neighbourhood = LookupNeighbourhood(neighbourhood_data)
    neighbours = neighbourhood.get_neighbours_up_to_distance(index, distance)
    assert neighbours == expected

    neighbours = neighbourhood.get_neighbours_up_to_distance(index, distance, include_center=True)
    assert neighbours == expected_with_include_center

    neighbourhood = LookupNeighbourhood(neighbourhood_data, include_center=True)
    neighbours = neighbourhood.get_neighbours_up_to_distance(index, distance)
    assert neighbours == expected_with_include_center

    neighbours = neighbourhood.get_neighbours_up_to_distance(index, distance, include_center=False)
    assert neighbours == expected



"""
OSM PBF Loader.

This module contains loader capable of loading OpenStreetMap features from `*.osm.pbf` files.
"""

from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

import geopandas as gpd
from shapely.geometry.base import BaseGeometry

from srai._optional import import_optional_dependencies
from srai.constants import GEOMETRY_COLUMN, WGS84_CRS
from srai.loaders.osm_loaders._base import OSMLoader
from srai.loaders.osm_loaders.filters import GroupedOsmTagsFilter, OsmTagsFilter

if TYPE_CHECKING:
    from quackosm import PbfFileReader
    from quackosm.osm_extracts import OsmExtractSource


class OSMPbfLoader(OSMLoader):
    """
    OSMPbfLoader.

    OSM(OpenStreetMap)[1] PBF(Protocolbuffer Binary Format)[2] loader is a loader
    capable of loading OSM features from a PBF file. It filters features based on OSM tags[3]
    in form of key:value pairs, that are used by OSM users to give meaning to geometries.

    This loader uses `PbfFileReader` from the `QuackOSM`[3] library.
    It utilizes the `duckdb`[4] engine with `spatial`[5] extension
    capable of parsing an `*.osm.pbf` file.

    Additionally, it can download a pbf file extract for a given area using different sources.

    References:
        1. https://www.openstreetmap.org/
        2. https://wiki.openstreetmap.org/wiki/PBF_Format
        3. https://github.com/kraina-ai/quackosm
        4. https://duckdb.org/
        5. https://github.com/duckdb/duckdb_spatial
    """

    def __init__(
        self,
        pbf_file: Optional[Union[str, Path]] = None,
        download_source: "OsmExtractSource" = "geofabrik",
        download_directory: Union[str, Path] = "files",
    ) -> None:
        """
        Initialize OSMPbfLoader.

        Args:
            pbf_file (Union[str, Path], optional): Downloaded `*.osm.pbf` file to be used by
                the loader. If not provided, it will be automatically downloaded for a given area.
                Defaults to None.
            download_source (OsmExtractSource, optional): Source to use when downloading PBF files.
                Can be one of: `any`, `geofabrik`, `osmfr`, `bbbike`.
                Defaults to "any".
            download_directory (Union[str, Path], optional): Directory where to save the downloaded
                `*.osm.pbf` files. Ignored if `pbf_file` is provided. Defaults to "files".
        """
        import_optional_dependencies(dependency_group="osm", modules=["quackosm"])
        self.pbf_file = pbf_file
        self.download_source = download_source
        self.download_directory = download_directory

    def load(
        self,
        area: Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame],
        tags: Union[OsmTagsFilter, GroupedOsmTagsFilter],
        ignore_cache: bool = False,
        explode_tags: bool = True,
        keep_all_tags: bool = False,
    ) -> gpd.GeoDataFrame:
        """
        Load OSM features with specified tags for a given area from an `*.osm.pbf` file.

        The loader will use provided `*.osm.pbf` file, or download extracts
        automatically. Later it will parse and filter features from files
        using `PbfFileReader` from `QuackOSM` library. It will return a GeoDataFrame
        containing the `geometry` column and columns for tag keys.

        Note: Some key/value pairs might be missing from the resulting GeoDataFrame,
            simply because there are no such objects in the given area.

        Args:
            area (Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame]):
                Area for which to download objects.
            tags (Union[OsmTagsFilter, GroupedOsmTagsFilter]): A dictionary
                specifying which tags to download.
                The keys should be OSM tags (e.g. `building`, `amenity`).
                The values should either be `True` for retrieving all objects with the tag,
                string for retrieving a single tag-value pair
                or list of strings for retrieving all values specified in the list.
                `tags={'leisure': 'park}` would return parks from the area.
                `tags={'leisure': 'park, 'amenity': True, 'shop': ['bakery', 'bicycle']}`
                would return parks, all amenity types, bakeries and bicycle shops.
            ignore_cache: (bool, optional): Whether to ignore precalculated geoparquet files or not.
                Defaults to False.
            explode_tags: (bool, optional): Whether to split OSM tags into multiple columns or keep
                them in a single dict. Defaults to True.
            keep_all_tags: (bool, optional): Whether to keep all tags related to the element,
                or return only those defined in the `tags_filter`. When True, will override
                the optional grouping defined in the `tags_filter`. Defaults to False.

        Raises:
            ValueError: If PBF file is expected to be downloaded and provided geometries
                aren't shapely.geometry.Polygons.

        Returns:
            gpd.GeoDataFrame: Downloaded features as a GeoDataFrame.
        """
        area_wgs84 = self._prepare_area_gdf(area)

        pbf_reader = self._get_pbf_file_reader(area_wgs84, tags)

        if self.pbf_file is not None:
            features_gdf = pbf_reader.get_features_gdf(
                file_paths=self.pbf_file,
                keep_all_tags=keep_all_tags,
                explode_tags=explode_tags,
                ignore_cache=ignore_cache,
            )
        else:
            features_gdf = pbf_reader.get_features_gdf_from_geometry(
                keep_all_tags=keep_all_tags, explode_tags=explode_tags, ignore_cache=ignore_cache
            )

        features_gdf = features_gdf.set_crs(WGS84_CRS)

        features_columns = [
            column
            for column in features_gdf.columns
            if column != GEOMETRY_COLUMN and features_gdf[column].notnull().any()
        ]
        features_gdf = features_gdf[[GEOMETRY_COLUMN, *sorted(features_columns)]]

        return features_gdf

    def load_to_geoparquet(
        self,
        area: Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame],
        tags: Union[OsmTagsFilter, GroupedOsmTagsFilter],
        ignore_cache: bool = False,
        explode_tags: bool = True,
        keep_all_tags: bool = False,
    ) -> Path:
        """
        Load OSM features with specified tags for a given area and save it to geoparquet file.

        Args:
            area (Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame]):
                Area for which to download objects.
            tags (Union[OsmTagsFilter, GroupedOsmTagsFilter]): A dictionary
                specifying which tags to download.
                The keys should be OSM tags (e.g. `building`, `amenity`).
                The values should either be `True` for retrieving all objects with the tag,
                string for retrieving a single tag-value pair
                or list of strings for retrieving all values specified in the list.
                `tags={'leisure': 'park}` would return parks from the area.
                `tags={'leisure': 'park, 'amenity': True, 'shop': ['bakery', 'bicycle']}`
                would return parks, all amenity types, bakeries and bicycle shops.
            ignore_cache: (bool, optional): Whether to ignore precalculated geoparquet files or not.
                Defaults to False.
            explode_tags: (bool, optional): Whether to split OSM tags into multiple columns or keep
                them in a single dict. Defaults to True.
            keep_all_tags: (bool, optional): Whether to keep all tags related to the element,
                or return only those defined in the `tags_filter`. When True, will override
                the optional grouping defined in the `tags_filter`. Defaults to False.

        Returns:
            Path: Path to the saved GeoParquet file.
        """
        area_wgs84 = self._prepare_area_gdf(area)

        pbf_reader = self._get_pbf_file_reader(area_wgs84, tags)

        geoparquet_file_path: Path

        if self.pbf_file is not None:
            geoparquet_file_path = pbf_reader.convert_pbf_to_gpq(
                pbf_path=self.pbf_file,
                keep_all_tags=keep_all_tags,
                explode_tags=explode_tags,
                ignore_cache=ignore_cache,
            )
        else:
            geoparquet_file_path = pbf_reader.convert_geometry_filter_to_gpq(
                keep_all_tags=keep_all_tags, explode_tags=explode_tags, ignore_cache=ignore_cache
            )

        return geoparquet_file_path

    def _get_pbf_file_reader(
        self, area_wgs84: gpd.GeoDataFrame, tags: Union[OsmTagsFilter, GroupedOsmTagsFilter]
    ) -> "PbfFileReader":
        from quackosm import PbfFileReader
        from quackosm.osm_extracts import OsmExtractSource

        pbf_reader = PbfFileReader(
            tags_filter=tags,
            geometry_filter=area_wgs84.unary_union,
            working_directory=self.download_directory,
            osm_extract_source=OsmExtractSource(self.download_source),
        )
        return pbf_reader



"""
Embedding model for gtfs2vec.

This module contains embedding model from gtfs2vec paper [1].

References:
    1. https://doi.org/10.1145/3486640.3491392
"""

from typing import TYPE_CHECKING, Any

from srai._optional import import_optional_dependencies
from srai.embedders import Model

if TYPE_CHECKING:  # pragma: no cover
    import torch


class GTFS2VecModel(Model):
    """Autoencoder based embedding model for gtfs2vec."""

    def __init__(
        self,
        n_features: int,
        n_hidden: int = 48,
        n_embed: int = 64,
    ) -> None:
        """
        Init GTFS2VecModel.

        Args:
            n_features (int): Number of features.
            n_hidden (int, optional): Number of hidden units. Defaults to 48.
            n_embed (int, optional): Embedding size. Defaults to 64.
        """
        super().__init__()
        import_optional_dependencies(
            dependency_group="torch", modules=["torch", "pytorch_lightning"]
        )
        from torch import nn

        self.n_features = n_features
        self.n_hidden = n_hidden
        self.n_embed = n_embed
        self.encoder = nn.Sequential(
            nn.Linear(n_features, n_hidden), nn.ReLU(), nn.Linear(n_hidden, n_embed)
        )
        self.decoder = nn.Sequential(
            nn.Linear(n_embed, n_hidden), nn.ReLU(), nn.Linear(n_hidden, n_features)
        )

    def forward(self, x: "torch.Tensor") -> "torch.Tensor":
        """
        Forward pass.

        Args:
            x (torch.Tensor): Input tensor.
        """
        embedding: "torch.Tensor" = self.encoder(x)
        return embedding

    def configure_optimizers(self) -> "torch.optim.Optimizer":
        """Configure optimizer."""
        import torch

        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
        return optimizer

    def training_step(self, batch: "torch.Tensor", batch_idx: Any) -> "torch.Tensor":
        """
        Training step.

        Args:
            batch (torch.Tensor): Batch.
            batch_idx (Any): Batch index.
        """
        from torch.nn import functional as F

        x = batch
        z = self.encoder(x)
        x_hat = self.decoder(z)
        loss = F.mse_loss(x_hat, x)
        self.log("train_loss", loss)
        return loss



"""
Spatial Representations for Artificial Intelligence.

srai is a Python module integrating many geo-related algorithms in a single package with unified
API. It aims to provide simple and efficient solutions to geospatial problems that are accessible to
everybody and reusable in various contexts where geospatial data can be used.

See
https://kraina-ai.github.io/srai/
for complete documentation.
"""

__version__ = "0.7.3"



"""Tests for OSMWayLoader."""

import pickle as pkl
from collections.abc import Sequence
from contextlib import nullcontext as does_not_raise
from functools import partial
from pathlib import Path
from typing import Any, Optional, Union
from unittest import TestCase

import geopandas as gpd
import numpy as np
import osmnx
import pandas as pd
import pytest
import shapely.geometry as shpg
from parametrization import Parametrization as P
from pytest_check import check
from pytest_mock import MockerFixture

from srai.constants import GEOMETRY_COLUMN, WGS84_CRS
from srai.exceptions import LoadedDataIsEmptyException
from srai.loaders import OSMNetworkType, OSMWayLoader

ut = TestCase()


@pytest.fixture  # type: ignore
def empty_area_gdf() -> gpd.GeoDataFrame:
    """Get a gdf with no geometry."""
    return gpd.GeoDataFrame(crs=WGS84_CRS, geometry=[])


@pytest.fixture  # type: ignore
def no_crs_area_gdf() -> gpd.GeoDataFrame:
    """Get an example area gdf without crs defined."""
    polygon = shpg.Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])
    gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [polygon]})
    return gdf


@pytest.fixture  # type: ignore
def bad_geometry_area_gdf() -> gpd.GeoDataFrame:
    """Get an example area gdf with unsupported geometry (not Polygon or Multipolygon)."""
    point = shpg.Point(0, 0)
    gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [point]}, crs=WGS84_CRS)
    return gdf


@pytest.fixture  # type: ignore
def empty_polygon_area_gdf() -> gpd.GeoDataFrame:
    """Get an example area gdf where there is no road infrastructure."""
    polygon = shpg.Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])
    gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [polygon]}, crs=WGS84_CRS)
    return gdf


@pytest.fixture  # type: ignore
def first_polygon_area_gdf() -> gpd.GeoDataFrame:
    """Get an example area gdf with one polygon."""
    polygon = shpg.Polygon(
        [
            (17.1005309, 51.1100158),
            (17.1020436, 51.1100427),
            (17.1021938, 51.1082509),
            (17.1006274, 51.1081027),
            (17.1005201, 51.1099956),
        ]
    )

    gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [polygon]}, crs=WGS84_CRS)
    return gdf


@pytest.fixture  # type: ignore
def second_polygon_area_gdf() -> gpd.GeoDataFrame:
    """Get an example area gdf with one polygon."""
    polygon = shpg.Polygon(
        [
            (17.0994473, 51.1084126),
            (17.1023226, 51.1086551),
            (17.1023333, 51.1076312),
            (17.0994473, 51.1083722),
        ]
    )

    gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [polygon]}, crs=WGS84_CRS)
    return gdf


@pytest.fixture  # type: ignore
def multiple_polygons_overlapping_area_gdf(
    first_polygon_area_gdf: gpd.GeoDataFrame, second_polygon_area_gdf: gpd.GeoDataFrame
) -> gpd.GeoDataFrame:
    """Get an example area gdf with two polygons."""
    return pd.concat([first_polygon_area_gdf, second_polygon_area_gdf], axis=0)


@pytest.fixture  # type: ignore
def multipolygons_area_gdf() -> gpd.GeoDataFrame:
    """Get an example area gdf with a multipolygon."""
    polygon1 = shpg.Polygon(
        [
            (17.1005309, 51.1100158),
            (17.1020436, 51.1100427),
            (17.1021938, 51.1082509),
            (17.1006274, 51.1081027),
            (17.1005201, 51.1099956),
        ]
    )
    polygon2 = shpg.Polygon(
        [
            (17.0997584, 51.1049434),
            (17.0995009, 51.1044112),
            (17.1003485, 51.1043910),
            (17.0997584, 51.1049434),
        ]
    )
    multipolygon = shpg.MultiPolygon([polygon1, polygon2])
    return gpd.GeoDataFrame(geometry=[multipolygon], crs=WGS84_CRS)


@pytest.fixture  # type: ignore
def valid_and_empty_polygons_area_gdf(
    first_polygon_area_gdf: gpd.GeoDataFrame, empty_polygon_area_gdf: gpd.GeoDataFrame
) -> gpd.GeoDataFrame:
    """Get an example area gdf with one valid polygon and one without road infrastructure."""
    return pd.concat([first_polygon_area_gdf, empty_polygon_area_gdf], axis=0)


@P.parameters(  # type: ignore
    "area_gdf_fixture", "expected_result", "loader_params", "file_name", "expectation"
)
@P.case(  # type: ignore
    "Raise when no geometry", "empty_area_gdf", None, None, None, pytest.raises(ValueError)
)
@P.case(  # type: ignore
    "Raise when no CRS",
    "no_crs_area_gdf",
    None,
    None,
    None,
    pytest.raises(ValueError),
)
@P.case(  # type: ignore
    "Raise when invalid geometry",
    "bad_geometry_area_gdf",
    None,
    None,
    "type_error",
    pytest.raises(TypeError),
)
@P.case(  # type: ignore
    "Raise when no road infrastructure",
    "empty_polygon_area_gdf",
    None,
    None,
    "empty_overpass_response",
    pytest.raises(LoadedDataIsEmptyException),
)
@P.case(  # type: ignore
    "Return infrastructure when single polygon",
    "first_polygon_area_gdf",
    (7, 6),
    None,
    "graph_1",
    does_not_raise(),
)
@P.case(  # type: ignore
    "Return infrastructure when multiple overlapping polygons",
    "multiple_polygons_overlapping_area_gdf",
    (7, 7),  # FIXME: should be (7, 6)
    None,
    "graph_1",
    does_not_raise(),
)
@P.case(  # type: ignore
    "Return infrastructure when node without any edges",  # FIXME: shouldn't have a node w/o an edge
    "multiple_polygons_overlapping_area_gdf",
    (9, 7),
    {"network_type": OSMNetworkType.BIKE, "contain_within_area": True},
    "graph_2",
    does_not_raise(),
)
@P.case(  # type: ignore
    "Return infrastructure when multipolygon",
    "multipolygons_area_gdf",
    (11, 9),
    None,
    "graph_3",
    does_not_raise(),
)
@P.case(  # type: ignore
    "Return infrastructure when correct polygon and polygon with no road infrastructure",
    "valid_and_empty_polygons_area_gdf",
    (7, 6),
    None,
    "graph_1",
    does_not_raise(),
)
@P.case(  # type: ignore
    "Return infrastructure without preprocessing",
    "first_polygon_area_gdf",
    (7, 6),
    {"network_type": OSMNetworkType.DRIVE, "preprocess": False},
    "graph_1",
    does_not_raise(),
)
@P.case(  # type: ignore
    "Return infrastructure in long format",
    "first_polygon_area_gdf",
    (7, 6),
    {"network_type": OSMNetworkType.DRIVE, "wide": False},
    "graph_1",
    does_not_raise(),
)
@P.case(  # type: ignore
    "Return infrastructure when network_type supplied as a string",
    "first_polygon_area_gdf",
    (7, 6),
    {"network_type": "drive", "wide": False},
    "graph_4",
    does_not_raise(),
)
def test_contract(
    area_gdf_fixture: str,
    expected_result: Optional[tuple[int, int]],
    loader_params: Optional[dict[str, Any]],
    file_name: Optional[str],
    expectation,
    request: pytest.FixtureRequest,
    mocker: MockerFixture,
):
    """Test `OSMWayLoader.load`'s contract."""
    if not loader_params:
        loader_params = {"network_type": OSMNetworkType.DRIVE}

    loader = OSMWayLoader(**loader_params)

    area_gdf = request.getfixturevalue(area_gdf_fixture)
    nodes_expected_len, edges_expected_len = expected_result or (None, None)

    def patched_graph_from_polygon(f_name: str, *args, **kwargs) -> Any:  # type: ignore
        if f_name == "type_error":
            raise TypeError
        elif f_name == "empty_overpass_response":
            from packaging import version

            osmnx_new_api = version.parse(osmnx.__version__) >= version.parse("1.6.0")

            if osmnx_new_api:
                raise osmnx._errors.InsufficientResponseError
            else:
                raise osmnx._errors.EmptyOverpassResponse

        files_path = Path(__file__).parent / "test_files"
        file_name = f_name + ".pkl"
        file_path = files_path / file_name

        with file_path.open("rb") as f:
            res = pkl.load(f)

        return res

    with expectation:
        mocker.patch(
            "osmnx.graph_from_polygon",
            wraps=partial(patched_graph_from_polygon, file_name),
        )
        nodes, edges = loader.load(area_gdf)

        check.equal(len(nodes), nodes_expected_len)
        check.equal(len(edges), edges_expected_len)
        check.is_in(GEOMETRY_COLUMN, nodes.columns)
        check.is_in(GEOMETRY_COLUMN, edges.columns)


@P.parameters("column_name", "input", "expected")  # type: ignore
@P.case(
    "Return None when `oneway` is missing",
    "oneway",
    ("", "none", "None", np.nan, "nan", "NaN", None),
    None,
)  # type: ignore
@P.case(
    "Return None when `lanes` is missing",
    "lanes",
    ("", "none", "None", np.nan, "nan", "NaN", None),
    None,
)  # type: ignore
@P.case(
    "Return None when `highway` is missing",
    "highway",
    ("", "none", "None", np.nan, "nan", "NaN", None),
    None,
)  # type: ignore
@P.case(
    "Test lanes",
    "lanes",
    [-1, 0, 1, 14, 15, 16, 17, 100, 3.7, "a", "2"],
    [-1, 0, 1, 14, 15, 15, 15, 15, 3, None, 2],
)  # type: ignore
@P.case(  # type: ignore
    "Test maxspeed",
    "maxspeed",
    [
        -1,
        0,
        1,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        14,
        15,
        16,
        17,
        44,
        45,
        46,
        199,
        200,
        201,
        300,
        6.7,
        "3",
        "4.5 km/h",
        "4.5km/h",
        "3mph",
        "4 mph",
        "signals",
        "variable",
        "CZ:pedestrian_zone",
    ],
    [
        0,
        0,
        5,
        5,
        5,
        7,
        7,
        10,
        10,
        10,
        15,
        15,
        15,
        20,
        20,
        40,
        40,
        50,
        200,
        200,
        200,
        200,
        7,
        5,
        5,
        5,
        5,
        7,
        None,
        None,
        20,
    ],
)
@P.case(
    "Test width",
    "width",
    [
        -1.3,
        -1,
        0,
        1,
        2,
        1.0,
        4.4,
        4.5,
        4.6,
        29,
        30,
        31,
        40,
        100,
        "4.6",
        "12m",
        "12 m",
        "12meter",
        "12 meter",
        "130.4'",
        "130.4 '",
        "25.4ft",
        "25.4 ft",
    ],
    [
        -1.5,
        -1.0,
        0.0,
        1.0,
        2.0,
        1.0,
        4.5,
        4.5,
        4.5,
        29.0,
        30.0,
        30.0,
        30.0,
        30.0,
        4.5,
        12.0,
        12.0,
        12.0,
        12.0,
        3.5,
        3.5,
        7.5,
        7.5,
    ],
)  # type: ignore
def test_preprocessing(
    column_name: str, input: Union[Any, Sequence[Any]], expected: Union[Any, Sequence[Any]]
) -> None:
    """Test `OSMWayLoader._sanitize_and_normalize()` preprocessing."""
    loader = OSMWayLoader(network_type=OSMNetworkType.DRIVE)

    input = list(input) if isinstance(input, Sequence) else [input]
    expected = list(expected) if isinstance(expected, Sequence) else [expected]

    if len(expected) == 1:
        expected = expected * len(input)

    if len(expected) != len(input):
        raise ValueError(
            f"Mismatch in length between input {len(input)} and expected {len(expected)}."
        )

    for x, y in zip(input, expected):
        result = loader._sanitize_and_normalize(x, column_name)
        check.equal(result, str(y))



"""
This module contains joiners, used to match spatial data with regions used in analysis.

All methods for spatial embedding (available in `srai`) are designed to operate on regions, not on
individual spatial features. This means that we need to match spatial features with our given
regions. This can be done in different ways, all of which are available under a common `Joiner`
interface.
"""

from ._base import Joiner
from .intersection_joiner import IntersectionJoiner

__all__ = [
    "Joiner",
    "IntersectionJoiner",
]



"""ContextualCountEmbedder tests."""

from contextlib import nullcontext as does_not_raise
from typing import TYPE_CHECKING, Any, Union

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from parametrization import Parametrization as P

from srai.constants import REGIONS_INDEX
from srai.embedders import ContextualCountEmbedder
from srai.loaders.osm_loaders.filters import OsmTagsFilter
from srai.neighbourhoods import H3Neighbourhood

if TYPE_CHECKING:  # pragma: no cover
    import geopandas as gpd


def _create_features_dataframe(data: dict[str, Any]) -> pd.DataFrame:
    return pd.DataFrame(data).set_index(REGIONS_INDEX)


@pytest.fixture  # type: ignore
def expected_embedding_df_squashed_distance_0() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output.

    Count without subcategories. Squashed features, distance 0.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure": [0, 1, 1],
            "amenity": [1, 0, 1],
        }
    )


@pytest.fixture  # type: ignore
def expected_embedding_df_squashed_distance_1() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output.

    Count without subcategories. Squashed features, distance 1+.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure": [0.25, 1.125, 1.125],
            "amenity": [1.125, 0.25, 1.125],
        },
    )


@pytest.fixture  # type: ignore
def expected_embedding_df_concatenated_distance_0() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output.

    Count without subcategories. Concatenated features, distance 0.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure_0": [0, 1, 1],
            "amenity_0": [1, 0, 1],
        }
    )


@pytest.fixture  # type: ignore
def expected_embedding_df_concatenated_distance_1() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output.

    Count without subcategories. Concatenated features, distance 1.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure_0": [0, 1, 1],
            "amenity_0": [1, 0, 1],
            "leisure_1": [1, 0.5, 0.5],
            "amenity_1": [0.5, 1, 0.5],
        },
    )


@pytest.fixture  # type: ignore
def expected_embedding_df_concatenated_distance_2() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output.

    Count without subcategories. Concatenated features, distance 2.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure_0": [0, 1, 1],
            "amenity_0": [1, 0, 1],
            "leisure_1": [1, 0.5, 0.5],
            "amenity_1": [0.5, 1, 0.5],
            "leisure_2": [0, 0, 0],
            "amenity_2": [0, 0, 0],
        },
    )


@pytest.fixture  # type: ignore
def expected_subcategories_embedding_df_squashed_distance_0() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output.

    Count with subcategories. Squashed features, distance 0.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure_adult_gaming_centre": [0, 0, 1],
            "leisure_playground": [0, 1, 0],
            "amenity_pub": [1, 0, 1],
        },
    )


@pytest.fixture  # type: ignore
def expected_subcategories_embedding_df_squashed_distance_1() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output.

    Count with subcategories. Squashed features, distance 1+.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure_adult_gaming_centre": [0.125, 0.125, 1],
            "leisure_playground": [0.125, 1, 0.125],
            "amenity_pub": [1.125, 0.25, 1.125],
        },
    )


@pytest.fixture  # type: ignore
def expected_subcategories_embedding_df_concatenated_distance_0() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output.

    Count with subcategories. Concatenated features, distance 0.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure_adult_gaming_centre_0": [0, 0, 1],
            "leisure_playground_0": [0, 1, 0],
            "amenity_pub_0": [1, 0, 1],
        },
    )


@pytest.fixture  # type: ignore
def expected_subcategories_embedding_df_concatenated_distance_1() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output.

    Count with subcategories. Concatenated features, distance 1.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure_adult_gaming_centre_0": [0, 0, 1],
            "leisure_playground_0": [0, 1, 0],
            "amenity_pub_0": [1, 0, 1],
            "leisure_adult_gaming_centre_1": [0.5, 0.5, 0],
            "leisure_playground_1": [0.5, 0, 0.5],
            "amenity_pub_1": [0.5, 1, 0.5],
        },
    )


@pytest.fixture  # type: ignore
def expected_subcategories_embedding_df_concatenated_distance_2() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output.

    Count with subcategories. Concatenated features, distance 2.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure_adult_gaming_centre_0": [0, 0, 1],
            "leisure_playground_0": [0, 1, 0],
            "amenity_pub_0": [1, 0, 1],
            "leisure_adult_gaming_centre_1": [0.5, 0.5, 0],
            "leisure_playground_1": [0.5, 0, 0.5],
            "amenity_pub_1": [0.5, 1, 0.5],
            "leisure_adult_gaming_centre_2": [0, 0, 0],
            "leisure_playground_2": [0, 0, 0],
            "amenity_pub_2": [0, 0, 0],
        },
    )


@pytest.fixture  # type: ignore
def expected_feature_names() -> list[str]:
    """Get expected feature names for ContextualCountEmbedder."""
    expected_feature_names = ["amenity_parking", "leisure_park", "amenity_pub"]
    return expected_feature_names


@pytest.fixture  # type: ignore
def osm_tags_filter() -> OsmTagsFilter:
    """Get osm tags filter for CountEmbedder."""
    tags_filter: OsmTagsFilter = {
        "amenity": ["parking", "pub"],
        "leisure": "park",
    }
    return tags_filter


@pytest.fixture  # type: ignore
def specified_features_expected_embedding_df_squashed_empty() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output for the case with specified features.

    Count without subcategories. Squashed features, distance 0+.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "amenity_parking": [0, 0, 0],
            "leisure_park": [0, 0, 0],
            "amenity_pub": [0, 0, 0],
        }
    )


@pytest.fixture  # type: ignore
def specified_features_expected_subcategories_embedding_df_squashed_distance_0() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output for the case with specified features.

    Count with subcategories. Squashed features, distance 0.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "amenity_parking": [0, 0, 0],
            "leisure_park": [0, 0, 0],
            "amenity_pub": [1, 0, 1],
        }
    )


@pytest.fixture  # type: ignore
def specified_features_expected_subcategories_embedding_df_squashed_distance_1() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output for the case with specified features.

    Count with subcategories. Squashed features, distance 1+.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "amenity_parking": [0, 0, 0],
            "leisure_park": [0, 0, 0],
            "amenity_pub": [1.125, 0.25, 1.125],
        }
    )


@pytest.fixture  # type: ignore
def specified_features_expected_embedding_df_concatenated_distance_0() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output for the case with specified features.

    Count without subcategories. Concatenated features, distance 0.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "amenity_parking_0": [0, 0, 0],
            "leisure_park_0": [0, 0, 0],
            "amenity_pub_0": [0, 0, 0],
        }
    )


@pytest.fixture  # type: ignore
def specified_features_expected_embedding_df_concatenated_distance_1() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output for the case with specified features.

    Count without subcategories. Concatenated features, distance 1.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "amenity_parking_0": [0, 0, 0],
            "leisure_park_0": [0, 0, 0],
            "amenity_pub_0": [0, 0, 0],
            "amenity_parking_1": [0, 0, 0],
            "leisure_park_1": [0, 0, 0],
            "amenity_pub_1": [0, 0, 0],
        }
    )


@pytest.fixture  # type: ignore
def specified_features_expected_embedding_df_concatenated_distance_2() -> pd.DataFrame:
    """
    Get expected ContextualCountEmbedder output for the case with specified features.

    Count without subcategories. Concatenated features, distance 2.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "amenity_parking_0": [0, 0, 0],
            "leisure_park_0": [0, 0, 0],
            "amenity_pub_0": [0, 0, 0],
            "amenity_parking_1": [0, 0, 0],
            "leisure_park_1": [0, 0, 0],
            "amenity_pub_1": [0, 0, 0],
            "amenity_parking_2": [0, 0, 0],
            "leisure_park_2": [0, 0, 0],
            "amenity_pub_2": [0, 0, 0],
        }
    )


@pytest.fixture  # type: ignore
def specified_features_expected_subcategories_embedding_df_concatenated_distance_0() -> (
    pd.DataFrame
):
    """
    Get expected ContextualCountEmbedder output for the case with specified features.

    Count with subcategories. Concatenated features, distance 0.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "amenity_parking_0": [0, 0, 0],
            "leisure_park_0": [0, 0, 0],
            "amenity_pub_0": [1, 0, 1],
        }
    )


@pytest.fixture  # type: ignore
def specified_features_expected_subcategories_embedding_df_concatenated_distance_1() -> (
    pd.DataFrame
):
    """
    Get expected ContextualCountEmbedder output for the case with specified features.

    Count with subcategories. Concatenated features, distance 1.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "amenity_parking_0": [0, 0, 0],
            "leisure_park_0": [0, 0, 0],
            "amenity_pub_0": [1, 0, 1],
            "amenity_parking_1": [0, 0, 0],
            "leisure_park_1": [0, 0, 0],
            "amenity_pub_1": [0.5, 1, 0.5],
        }
    )


@pytest.fixture  # type: ignore
def specified_features_expected_subcategories_embedding_df_concatenated_distance_2() -> (
    pd.DataFrame
):
    """
    Get expected ContextualCountEmbedder output for the case with specified features.

    Count with subcategories. Concatenated features, distance 2.
    """
    return _create_features_dataframe(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "amenity_parking_0": [0, 0, 0],
            "leisure_park_0": [0, 0, 0],
            "amenity_pub_0": [1, 0, 1],
            "amenity_parking_1": [0, 0, 0],
            "leisure_park_1": [0, 0, 0],
            "amenity_pub_1": [0.5, 1, 0.5],
            "amenity_parking_2": [0, 0, 0],
            "leisure_park_2": [0, 0, 0],
            "amenity_pub_2": [0, 0, 0],
        }
    )


@P.parameters(
    "expected_embedding_fixture",
    "neighbourhood_distance",
    "concatenate_features",
    "count_subcategories",
    "expected_features_fixture",
)  # type: ignore
@P.case(  # type: ignore
    "Squashed features, distance 0, without subcategories",
    "expected_embedding_df_squashed_distance_0",
    0,
    False,
    False,
    None,
)
@P.case(  # type: ignore
    "Squashed features, distance 1, without subcategories",
    "expected_embedding_df_squashed_distance_1",
    1,
    False,
    False,
    None,
)
@P.case(  # type: ignore
    "Squashed features, distance 2, without subcategories",
    "expected_embedding_df_squashed_distance_1",
    2,
    False,
    False,
    None,
)
@P.case(  # type: ignore
    "Concatenated features, distance 0, without subcategories",
    "expected_embedding_df_concatenated_distance_0",
    0,
    True,
    False,
    None,
)
@P.case(  # type: ignore
    "Concatenated features, distance 1, without subcategories",
    "expected_embedding_df_concatenated_distance_1",
    1,
    True,
    False,
    None,
)
@P.case(  # type: ignore
    "Concatenated features, distance 2, without subcategories",
    "expected_embedding_df_concatenated_distance_2",
    2,
    True,
    False,
    None,
)
@P.case(  # type: ignore
    "Squashed features, distance 0, witt subcategories",
    "expected_subcategories_embedding_df_squashed_distance_0",
    0,
    False,
    True,
    None,
)
@P.case(  # type: ignore
    "Squashed features, distance 1, with subcategories",
    "expected_subcategories_embedding_df_squashed_distance_1",
    1,
    False,
    True,
    None,
)
@P.case(  # type: ignore
    "Squashed features, distance 2, with subcategories",
    "expected_subcategories_embedding_df_squashed_distance_1",
    2,
    False,
    True,
    None,
)
@P.case(  # type: ignore
    "Concatenated features, distance 0, with subcategories",
    "expected_subcategories_embedding_df_concatenated_distance_0",
    0,
    True,
    True,
    None,
)
@P.case(  # type: ignore
    "Concatenated features, distance 1, with subcategories",
    "expected_subcategories_embedding_df_concatenated_distance_1",
    1,
    True,
    True,
    None,
)
@P.case(  # type: ignore
    "Concatenated features, distance 2, with subcategories",
    "expected_subcategories_embedding_df_concatenated_distance_2",
    2,
    True,
    True,
    None,
)
@P.case(  # type: ignore
    "Squashed features, distance 0, without subcategories, specified features",
    "specified_features_expected_embedding_df_squashed_empty",
    0,
    False,
    False,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Squashed features, distance 1, without subcategories, specified features",
    "specified_features_expected_embedding_df_squashed_empty",
    1,
    False,
    False,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Squashed features, distance 2, without subcategories, specified features",
    "specified_features_expected_embedding_df_squashed_empty",
    2,
    False,
    False,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Squashed features, distance 0, with subcategories, specified features",
    "specified_features_expected_subcategories_embedding_df_squashed_distance_0",
    0,
    False,
    True,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Squashed features, distance 1, with subcategories, specified features",
    "specified_features_expected_subcategories_embedding_df_squashed_distance_1",
    1,
    False,
    True,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Squashed features, distance 2, with subcategories, specified features",
    "specified_features_expected_subcategories_embedding_df_squashed_distance_1",
    2,
    False,
    True,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Concatenated features, distance 0, without subcategories, specified features",
    "specified_features_expected_embedding_df_concatenated_distance_0",
    0,
    True,
    False,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Concatenated features, distance 1, without subcategories, specified features",
    "specified_features_expected_embedding_df_concatenated_distance_1",
    1,
    True,
    False,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Concatenated features, distance 2, without subcategories, specified features",
    "specified_features_expected_embedding_df_concatenated_distance_2",
    2,
    True,
    False,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Concatenated features, distance 0, with subcategories, specified features",
    "specified_features_expected_subcategories_embedding_df_concatenated_distance_0",
    0,
    True,
    True,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Concatenated features, distance 1, with subcategories, specified features",
    "specified_features_expected_subcategories_embedding_df_concatenated_distance_1",
    1,
    True,
    True,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Concatenated features, distance 2, with subcategories, specified features",
    "specified_features_expected_subcategories_embedding_df_concatenated_distance_2",
    2,
    True,
    True,
    "expected_feature_names",
)
@P.case(  # type: ignore
    "Squashed features, distance 0, without subcategories, specified osm tags filter",
    "expected_embedding_df_squashed_distance_0",
    0,
    False,
    False,
    "osm_tags_filter",
)
@P.case(  # type: ignore
    "Squashed features, distance 1, without subcategories, specified osm tags filter",
    "expected_embedding_df_squashed_distance_1",
    1,
    False,
    False,
    "osm_tags_filter",
)
@P.case(  # type: ignore
    "Squashed features, distance 2, without subcategories, specified osm tags filter",
    "expected_embedding_df_squashed_distance_1",
    2,
    False,
    False,
    "osm_tags_filter",
)
@P.case(  # type: ignore
    "Squashed features, distance 0, with subcategories, specified osm tags filter",
    "specified_features_expected_subcategories_embedding_df_squashed_distance_0",
    0,
    False,
    True,
    "osm_tags_filter",
)
@P.case(  # type: ignore
    "Squashed features, distance 1, with subcategories, specified osm tags filter",
    "specified_features_expected_subcategories_embedding_df_squashed_distance_1",
    1,
    False,
    True,
    "osm_tags_filter",
)
@P.case(  # type: ignore
    "Squashed features, distance 2, with subcategories, specified osm tags filter",
    "specified_features_expected_subcategories_embedding_df_squashed_distance_1",
    2,
    False,
    True,
    "osm_tags_filter",
)
@P.case(  # type: ignore
    "Concatenated features, distance 0, without subcategories, specified osm tags filter",
    "expected_embedding_df_concatenated_distance_0",
    0,
    True,
    False,
    "osm_tags_filter",
)
@P.case(  # type: ignore
    "Concatenated features, distance 1, without subcategories, specified osm tags filter",
    "expected_embedding_df_concatenated_distance_1",
    1,
    True,
    False,
    "osm_tags_filter",
)
@P.case(  # type: ignore
    "Concatenated features, distance 2, without subcategories, specified osm tags filter",
    "expected_embedding_df_concatenated_distance_2",
    2,
    True,
    False,
    "osm_tags_filter",
)
@P.case(  # type: ignore
    "Concatenated features, distance 0, with subcategories, specified osm tags filter",
    "specified_features_expected_subcategories_embedding_df_concatenated_distance_0",
    0,
    True,
    True,
    "osm_tags_filter",
)
@P.case(  # type: ignore
    "Concatenated features, distance 1, with subcategories, specified osm tags filter",
    "specified_features_expected_subcategories_embedding_df_concatenated_distance_1",
    1,
    True,
    True,
    "osm_tags_filter",
)
@P.case(  # type: ignore
    "Concatenated features, distance 2, with subcategories, specified osm tags filter",
    "specified_features_expected_subcategories_embedding_df_concatenated_distance_2",
    2,
    True,
    True,
    "osm_tags_filter",
)
def test_correct_embedding(
    expected_embedding_fixture: str,
    neighbourhood_distance: int,
    concatenate_features: bool,
    count_subcategories: bool,
    expected_features_fixture: Union[str, None],
    request: Any,
) -> None:
    """Test if ContextualCountEmbedder returns correct result with different parameters."""
    expected_output_features = (
        None
        if expected_features_fixture is None
        else request.getfixturevalue(expected_features_fixture)
    )
    gdf_regions: gpd.GeoDataFrame = request.getfixturevalue("gdf_regions")
    gdf_features: gpd.GeoDataFrame = request.getfixturevalue("gdf_features")
    gdf_joint: gpd.GeoDataFrame = request.getfixturevalue("gdf_joint")

    embedder = ContextualCountEmbedder(
        neighbourhood=H3Neighbourhood(),
        neighbourhood_distance=neighbourhood_distance,
        expected_output_features=expected_output_features,
        count_subcategories=count_subcategories,
        concatenate_vectors=concatenate_features,
    )
    embedding_df = embedder.transform(
        regions_gdf=gdf_regions, features_gdf=gdf_features, joint_gdf=gdf_joint
    )

    expected_result_df = request.getfixturevalue(expected_embedding_fixture)
    assert_frame_equal(
        embedding_df.sort_index(axis=1), expected_result_df.sort_index(axis=1), check_dtype=False
    )


def test_negative_nighbourhood_distance() -> None:
    """Test checks if negative neighbouthood distance is disallowed."""
    with pytest.raises(ValueError):
        ContextualCountEmbedder(neighbourhood=H3Neighbourhood(), neighbourhood_distance=-1)


@pytest.mark.parametrize(  # type: ignore
    "regions_fixture,features_fixture,joint_fixture,expected_features_fixture,expectation",
    [
        (
            "gdf_regions_empty",
            "gdf_features",
            "gdf_joint",
            None,
            does_not_raise(),
        ),
        (
            "gdf_regions",
            "gdf_features_empty",
            "gdf_joint",
            None,
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_features_empty",
            "gdf_joint",
            "expected_feature_names",
            does_not_raise(),
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_joint_empty",
            None,
            does_not_raise(),
        ),
    ],
)
@pytest.mark.parametrize("concatenate_features", [False, True])  # type: ignore
@pytest.mark.parametrize("count_subcategories", [False, True])  # type: ignore
@pytest.mark.parametrize("neighbourhood_distance", [0, 1, 2])  # type: ignore
def test_empty(
    regions_fixture: str,
    features_fixture: str,
    joint_fixture: str,
    concatenate_features: bool,
    count_subcategories: bool,
    neighbourhood_distance: int,
    expected_features_fixture: Union[str, None],
    expectation: Any,
    request: Any,
) -> None:
    """Test ContextualCountEmbedder handling of empty input data frames."""
    expected_output_features = (
        None
        if expected_features_fixture is None
        else request.getfixturevalue(expected_features_fixture)
    )
    embedder = ContextualCountEmbedder(
        neighbourhood=H3Neighbourhood(),
        neighbourhood_distance=neighbourhood_distance,
        expected_output_features=expected_output_features,
        count_subcategories=count_subcategories,
        concatenate_vectors=concatenate_features,
    )
    gdf_regions: gpd.GeoDataFrame = request.getfixturevalue(regions_fixture)
    gdf_features: gpd.GeoDataFrame = request.getfixturevalue(features_fixture)
    gdf_joint: gpd.GeoDataFrame = request.getfixturevalue(joint_fixture)

    with expectation:
        embedding = embedder.transform(gdf_regions, gdf_features, gdf_joint)
        assert len(embedding) == len(gdf_regions)
        assert embedding.index.name == gdf_regions.index.name
        if expected_output_features:
            assert len(embedding.columns) == len(expected_output_features) * (
                1 if not concatenate_features else 1 + neighbourhood_distance
            )

        assert (embedding == 0).all().all()


@pytest.mark.parametrize(  # type: ignore
    "regions_fixture,features_fixture,joint_fixture,expectation",
    [
        (
            "gdf_unnamed_single_index",
            "gdf_features",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_unnamed_single_index",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_unnamed_single_index",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_three_level_multi_index",
            pytest.raises(ValueError),
        ),
        (
            "gdf_incorrectly_named_single_index",
            "gdf_features",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_incorrectly_named_single_index",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
    ],
)
@pytest.mark.parametrize("concatenate_features", [False, True])  # type: ignore
@pytest.mark.parametrize("count_subcategories", [False, True])  # type: ignore
@pytest.mark.parametrize("neighbourhood_distance", [0, 1, 2])  # type: ignore
def test_incorrect_indexes(
    regions_fixture: str,
    features_fixture: str,
    joint_fixture: str,
    concatenate_features: bool,
    count_subcategories: bool,
    neighbourhood_distance: int,
    expectation: Any,
    request: Any,
) -> None:
    """Test if cannot embed with incorrect dataframe indexes."""
    regions_gdf: gpd.GeoDataFrame = request.getfixturevalue(regions_fixture)
    features_gdf: gpd.GeoDataFrame = request.getfixturevalue(features_fixture)
    joint_gdf: gpd.GeoDataFrame = request.getfixturevalue(joint_fixture)

    with expectation:
        ContextualCountEmbedder(
            neighbourhood=H3Neighbourhood(),
            count_subcategories=count_subcategories,
            concatenate_vectors=concatenate_features,
            neighbourhood_distance=neighbourhood_distance,
        ).transform(regions_gdf=regions_gdf, features_gdf=features_gdf, joint_gdf=joint_gdf)



"""Hex2Vec."""

from .embedder import Hex2VecEmbedder
from .model import Hex2VecModel
from .neighbour_dataset import NeighbourDataset, NeighbourDatasetItem

__all__ = ["Hex2VecEmbedder", "Hex2VecModel", "NeighbourDataset", "NeighbourDatasetItem"]



"""Voronoi regionalizer tests."""

from multiprocessing import cpu_count
from typing import Any, cast

import geopandas as gpd
import numpy as np
import pytest
from pymap3d import Ellipsoid
from shapely.geometry import Point, Polygon

from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS
from srai.geometry import merge_disjointed_gdf_geometries
from srai.regionalizers import VoronoiRegionalizer
from srai.regionalizers._spherical_voronoi import (
    _map_from_geocentric,
    _parse_multiprocessing_activation_threshold,
    _parse_num_of_multiprocessing_workers,
    generate_voronoi_regions,
)


def get_random_points(number_of_points: int, seed: int) -> list[Point]:
    """Get random points within WGS84 bounds sampled on a sphere."""
    vec = np.random.default_rng(seed=seed).standard_normal((3, number_of_points))
    vec /= np.linalg.norm(vec, axis=0)
    xi, yi, zi = vec

    unit_sphere_ellipsoid = Ellipsoid(
        semimajor_axis=1, semiminor_axis=1, name="Unit Sphere", model="Unit"
    )

    return [
        Point(_map_from_geocentric(_x, _y, _z, unit_sphere_ellipsoid))
        for _x, _y, _z in zip(xi, yi, zi)
    ]


def list_to_geodataframe(points: list[Point]) -> gpd.GeoDataFrame:
    """Wrap points into GeoDataFrame."""
    return gpd.GeoDataFrame(
        {GEOMETRY_COLUMN: points},
        crs=WGS84_CRS,
    )


def check_if_disjoint(gdf: gpd.GeoDataFrame) -> bool:
    """Check if provided geometries are disjoint."""
    duplicated_regions = (
        gdf.sjoin(gdf, predicate="overlaps").index.value_counts().loc[lambda x: x > 1]
    )
    duplicated_seeds_ids = duplicated_regions.index.to_list()
    return len(duplicated_seeds_ids) == 0


def check_if_seeds_match_regions(seeds: gpd.GeoDataFrame, regions: gpd.GeoDataFrame) -> bool:
    """Check if order of seeds and regions is the same."""
    seeds_index_column_name = f"{seeds.index.name or 'index'}_seeds"
    joined_data = regions.reset_index().sjoin(seeds, rsuffix="seeds")

    if len(joined_data) != len(regions):
        return False

    return cast(bool, (joined_data[REGIONS_INDEX] == joined_data[seeds_index_column_name]).all())


def test_empty_gdf_attribute_error(gdf_empty: gpd.GeoDataFrame) -> None:
    """Test checks if empty GeoDataFrames are disallowed."""
    with pytest.raises(AttributeError):
        VoronoiRegionalizer(seeds=gdf_empty)


def test_no_crs_gdf_value_error(
    gdf_earth_poles: gpd.GeoDataFrame, gdf_no_crs: gpd.GeoDataFrame
) -> None:
    """Test checks if GeoDataFrames without crs are disallowed."""
    with pytest.raises(ValueError):
        vr = VoronoiRegionalizer(seeds=gdf_earth_poles)
        vr.transform(gdf=gdf_no_crs)


def test_duplicate_seeds_value_error() -> None:
    """Test checks if duplicate points are disallowed."""
    with pytest.raises(ValueError):
        seeds_gdf = gpd.GeoDataFrame(
            {GEOMETRY_COLUMN: [Point(0, 0), Point(0, 0), Point(-1, -1), Point(2, 2)]},
            index=[1, 2, 3, 4],
            crs=WGS84_CRS,
        )
        VoronoiRegionalizer(seeds=seeds_gdf)


def test_seed_outside_earth_bounding_box_value_error() -> None:
    """Test checks if points outside Earth bounding box are disallowed."""
    with pytest.raises(ValueError):
        seeds = [Point(0, 0), Point(-1, -1), Point(2, 2), Point(200, 200)]
        VoronoiRegionalizer(seeds=seeds)


def test_single_seed_region() -> None:
    """Test checks if single seed is disallowed."""
    with pytest.raises(ValueError):
        seeds_gdf = gpd.GeoDataFrame(
            {GEOMETRY_COLUMN: [Point(0, 0)]},
            index=[1],
            crs=WGS84_CRS,
        )
        VoronoiRegionalizer(seeds=seeds_gdf)


def test_single_seed_algorithm_error() -> None:
    """Test checks if single seed is disallowed."""
    with pytest.raises(ValueError):
        generate_voronoi_regions(seeds=[Point(0, 0)], max_meters_between_points=10_000)


@pytest.mark.parametrize(  # type: ignore
    "num_of_multiprocessing_workers,expected_num_of_multiprocessing_workers",
    [
        (-1, cpu_count()),
        (0, 1),
        (1, 1),
        (cpu_count() - 1, cpu_count() - 1),
        (cpu_count(), cpu_count()),
    ],
)
def test_num_of_multiprocessing_workers(
    num_of_multiprocessing_workers: int, expected_num_of_multiprocessing_workers: int
) -> None:
    """Test checks if number of workers is parsed correctly."""
    assert (
        _parse_num_of_multiprocessing_workers(num_of_multiprocessing_workers)
        == expected_num_of_multiprocessing_workers
    )


@pytest.mark.parametrize(  # type: ignore
    "multiprocessing_activation_threshold,expected_multiprocessing_activation_threshold",
    [
        (None, 100),
        (0, 100),
        (1, 1),
        (100, 100),
        (1_000, 1_000),
        (10_000, 10_000),
    ],
)
def test_multiprocessing_activation_threshold(
    multiprocessing_activation_threshold: int,
    expected_multiprocessing_activation_threshold: int,
) -> None:
    """Test checks if multiprocessing activation threshold is parsed correctly."""
    assert (
        _parse_multiprocessing_activation_threshold(multiprocessing_activation_threshold)
        == expected_multiprocessing_activation_threshold
    )


@pytest.mark.parametrize(  # type: ignore
    "max_meters_between_points",
    [100_000, 10_000, 1_000, 500],
)
def test_regions_edge_resolution(
    max_meters_between_points: int,
    gdf_earth_poles: gpd.GeoDataFrame,
    earth_bbox: Polygon,
) -> None:
    """Test checks if regions with different resolution are generated correctly."""
    vr = VoronoiRegionalizer(
        seeds=gdf_earth_poles,
        max_meters_between_points=max_meters_between_points,
        multiprocessing_activation_threshold=6,
    )
    result_gdf = vr.transform()
    assert len(result_gdf.index) == 6
    assert check_if_seeds_match_regions(
        seeds=gdf_earth_poles, regions=result_gdf
    ), "Seeds don't match generated regions"
    assert result_gdf.geometry.unary_union.difference(
        earth_bbox
    ).is_empty, "Result doesn't cover bounding box"
    assert check_if_disjoint(result_gdf), "Result isn't disjoint"


@pytest.mark.parametrize("random_points", [10, 100, 1_000, 10_000])  # type: ignore
def test_multiple_seeds_regions(
    random_points: int,
    earth_bbox: Polygon,
) -> None:
    """Test checks if regions are generated correctly."""
    seed = np.random.default_rng().integers(100_000_000)
    seeds = get_random_points(random_points, seed)
    vr = VoronoiRegionalizer(seeds=seeds)
    result_gdf = vr.transform()
    assert len(result_gdf.index) == random_points
    assert check_if_seeds_match_regions(
        seeds=gpd.GeoDataFrame(geometry=seeds, crs=WGS84_CRS), regions=result_gdf
    ), f"Seeds don't match generated regions (seed: {seed})"
    assert result_gdf.geometry.unary_union.difference(
        earth_bbox
    ).is_empty, f"Result doesn't cover bounding box (seed: {seed})"
    assert check_if_disjoint(result_gdf), f"Result isn't disjoint (seed: {seed})"


def test_four_close_seed_region(gdf_earth_bbox: gpd.GeoDataFrame, earth_bbox: Polygon) -> None:
    """Test checks if four close seeds are properly evaluated."""
    seeds_gdf = gpd.GeoDataFrame(
        {
            GEOMETRY_COLUMN: [
                Point(17.014997869227177, 51.09919872601259),
                Point(16.935542631959215, 51.09380600286582),
                Point(16.900425, 51.1162552343),
                Point(16.932700, 51.166251),
            ]
        },
        index=[1, 2, 3, 4],
        crs=WGS84_CRS,
    )
    vr = VoronoiRegionalizer(seeds=seeds_gdf)
    result_gdf = vr.transform(gdf=gdf_earth_bbox)
    assert len(result_gdf.index) == 4
    assert check_if_seeds_match_regions(
        seeds=seeds_gdf, regions=result_gdf
    ), "Seeds don't match generated regions"
    assert result_gdf.geometry.unary_union.difference(
        earth_bbox
    ).is_empty, "Result doesn't cover bounding box"
    assert check_if_disjoint(result_gdf), "Result isn't disjoint"


def test_default_parameter(gdf_earth_poles: gpd.GeoDataFrame, earth_bbox: Polygon) -> None:
    """Test checks if regions are generated correctly with a default mask."""
    vr = VoronoiRegionalizer(seeds=gdf_earth_poles)
    result_gdf = vr.transform(gdf=None)
    assert len(result_gdf.index) == 6
    assert check_if_seeds_match_regions(
        seeds=gdf_earth_poles, regions=result_gdf
    ), "Seeds don't match generated regions"
    assert result_gdf.unary_union.difference(earth_bbox).is_empty
    assert check_if_disjoint(result_gdf), "Result isn't disjoint"


@pytest.mark.parametrize(  # type: ignore
    "gdf_fixture",
    [
        "gdf_multipolygon",
        "gdf_poland",
        "gdf_earth_bbox",
    ],
)
def test_clipping_parameter(
    gdf_fixture: str,
    request: Any,
) -> None:
    """Test checks if regions are clipped correctly with a provided mask."""
    gdf: gpd.GeoDataFrame = request.getfixturevalue(gdf_fixture)
    gdf_earth_poles: gpd.GeoDataFrame = request.getfixturevalue("gdf_earth_poles")
    vr = VoronoiRegionalizer(seeds=gdf_earth_poles)
    result_gdf = vr.transform(gdf=gdf)
    assert merge_disjointed_gdf_geometries(result_gdf).difference(gdf.iloc[0].geometry).is_empty



"""Filters."""

from ._typing import (
    GroupedOsmTagsFilter,
    OsmTagsFilter,
    merge_osm_tags_filter,
)
from .base_osm_groups import BASE_OSM_GROUPS_FILTER
from .geofabrik import GEOFABRIK_LAYERS
from .hex2vec import HEX2VEC_FILTER
from .popular import get_popular_tags

__all__ = [
    "GroupedOsmTagsFilter",
    "OsmTagsFilter",
    "merge_osm_tags_filter",
    "BASE_OSM_GROUPS_FILTER",
    "GEOFABRIK_LAYERS",
    "HEX2VEC_FILTER",
    "get_popular_tags",
]



"""
H3 regionalizer.

This module exposes Uber's H3 Hexagonal Hierarchical Geospatial Indexing System [1] as
a regionalizer.

Note:
    The default API [2] was chosen (basic_str) to ease the implementation.
    It may be beneficial to try the NumPy API for computationally-heavy work.

References:
    1. https://uber.github.io/h3-py/
    2. https://uber.github.io/h3-py/api_comparison
"""

import geopandas as gpd

from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS
from srai.h3 import h3_to_geoseries, shapely_geometry_to_h3
from srai.regionalizers import Regionalizer


class H3Regionalizer(Regionalizer):
    """
    H3 Regionalizer.

    H3 Regionalizer allows the given geometries to be divided
    into H3 cells - hexagons with pentagons as a very rare exception
    """

    def __init__(self, resolution: int, buffer: bool = True) -> None:
        """
        Init H3Regionalizer.

        Args:
            resolution (int): Resolution of the cells. See [1] for a full comparison.
            buffer (bool, optional): Whether to fully cover the geometries with
                H3 Cells (visible on the borders). Defaults to True.

        Raises:
            ValueError: If resolution is not between 0 and 15.

        References:
            1. https://h3geo.org/docs/core-library/restable/
        """
        if not (0 <= resolution <= 15):
            raise ValueError(f"Resolution {resolution} is not between 0 and 15.")

        self.resolution = resolution
        self.buffer = buffer

    def transform(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """
        Regionalize a given GeoDataFrame.

        Transforms given geometries into H3 cells of given resolution
        and optionally applies buffering.

        Args:
            gdf (gpd.GeoDataFrame): (Multi)Polygons to be regionalized.

        Returns:
            gpd.GeoDataFrame: H3 cells.

        Raises:
            ValueError: If provided GeoDataFrame has no crs defined.
        """
        gdf_wgs84 = gdf.to_crs(crs=WGS84_CRS)

        gdf_exploded = self._explode_multipolygons(gdf_wgs84)

        h3_indexes = list(
            set(
                shapely_geometry_to_h3(
                    gdf_exploded[GEOMETRY_COLUMN],
                    h3_resolution=self.resolution,
                    buffer=self.buffer,
                )
            )
        )
        gdf_h3 = gpd.GeoDataFrame(
            data={REGIONS_INDEX: h3_indexes},
            geometry=h3_to_geoseries(h3_indexes),
            crs=WGS84_CRS,
        ).set_index(REGIONS_INDEX)

        return gdf_h3.to_crs(gdf.crs)



"""Constants for hex2vec tests."""

ENCODER_SIZES = [32, 16]
TRAINER_KWARGS = {"max_epochs": 1, "accelerator": "cpu", "deterministic": True}

PREDEFINED_TEST_CASES = [
    {
        "test_case_name": "wro_9",
        "geocoding_name": "Wrocław, Poland",
        "root_region_index": "891e2040887ffff",
        "h3_res": 9,
        "radius": 7,
        "seed": 42,
    },
    {
        "test_case_name": "poz_8",
        "geocoding_name": "Poznań, Poland",
        "root_region_index": "881e24a125fffff",
        "h3_res": 8,
        "radius": 6,
        "seed": 5555,
    },
]



"""Base class for joiners."""

import abc

import geopandas as gpd


class Joiner(abc.ABC):
    """Abstract class for joiners."""

    @abc.abstractmethod
    def transform(
        self,
        regions: gpd.GeoDataFrame,
        features: gpd.GeoDataFrame,
    ) -> gpd.GeoDataFrame:  # pragma: no cover
        """
        Join features to regions.

        Args:
            regions (gpd.GeoDataFrame): regions with which features are joined
            features (gpd.GeoDataFrame): features to be joined
        Returns:
            GeoDataFrame with an intersection of regions and features, which contains
            a MultiIndex and optionally a geometry with the intersection
        """
        raise NotImplementedError



"""Conftest for loaders."""

from datetime import datetime
from typing import Any

import pandas as pd
import pytest
from pytest_mock import MockerFixture


@pytest.fixture  # type: ignore
def gtfs_validation_ok() -> pd.DataFrame:
    """Get GTFS validation result with no errors."""
    return pd.DataFrame(
        {
            "type": ["warning", "warning"],
            "message": ["test warning", "test warning"],
            "table": ["test_table", "test_table"],
            "rows": [[1, 2], [3, 4]],
        }
    )


@pytest.fixture  # type: ignore
def gtfs_validation_error() -> pd.DataFrame:
    """Get GTFS validation result with errors."""
    return pd.DataFrame(
        {
            "type": ["error", "error", "warning", "warning"],
            "message": ["test error", "test error", "test warning", "test warning"],
            "table": ["test_table", "test_table", "test_table", "test_table"],
            "rows": [[1, 2], [3, 4], [5, 6], [7, 8]],
        }
    )


@pytest.fixture  # type: ignore
def stop_time_series() -> pd.DataFrame:
    """Get mocked stop time series."""
    ts = pd.DataFrame.from_dict(
        {
            "datetime": pd.DatetimeIndex(
                [datetime(2022, 1, 2, 12, 0), datetime(2022, 1, 2, 13, 0)]
            ),
            "42": pd.Series([0, 2]),
            "76": pd.Series([12, 12]),
        }
    ).set_index("datetime")

    ts.columns = pd.MultiIndex.from_tuples(
        [("num_trips", "42"), ("num_trips", "76")], names=["indicator", "stop_id"]
    )

    return ts


@pytest.fixture  # type: ignore
def stop_times() -> pd.DataFrame:
    """Get mocked stop times."""
    return pd.DataFrame(
        {
            "trip_id": ["1", "1", "2", "2"],
            "arrival_time": ["12:00:00", "13:00:00", "12:00:00", "13:00:00"],
            "departure_time": ["12:00:00", "12:00:00", "13:00:00", "13:00:00"],
            "stop_id": ["42", "76", "42", "76"],
        }
    )


@pytest.fixture  # type: ignore
def trips() -> pd.DataFrame:
    """Get mocked trips."""
    return pd.DataFrame(
        {
            "trip_id": ["1", "2"],
            "trip_headsign": ["A", "B"],
        }
    )


@pytest.fixture  # type: ignore
def stops() -> pd.DataFrame:
    """Get mocked stops."""
    return pd.DataFrame(
        {
            "stop_id": ["42", "76"],
            "stop_lat": [51.198083, 51.107133],
            "stop_lon": [16.905892, 17.019394],
        }
    )


@pytest.fixture  # type: ignore
def feed(
    mocker: MockerFixture,
    stop_time_series: pd.DataFrame,
    stop_times: pd.DataFrame,
    trips: pd.DataFrame,
    stops: pd.DataFrame,
) -> Any:
    """Get mocked feed."""
    feed_mock = mocker.MagicMock()
    feed_mock.configure_mock(
        **{
            "get_first_week.return_value": ["", "", "20220102"],
            "compute_stop_time_series.return_value": stop_time_series,
            "stop_times": stop_times,
            "trips": trips,
            "stops": stops,
        }
    )
    return feed_mock



"""
Spherical voronoi utils.

This module contains spherical voronoi implementation based on SphericalVoronoi function from scipy
library.
"""

import hashlib
import warnings
from collections.abc import Hashable
from contextlib import suppress
from functools import partial
from math import ceil
from multiprocessing import cpu_count
from typing import Optional, Union, cast

import geopandas as gpd
import numpy as np
import numpy.typing as npt
from haversine import haversine
from pymap3d import Ellipsoid, geodetic2ecef
from scipy.spatial import SphericalVoronoi, geometric_slerp
from shapely.geometry import (
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPolygon,
    Point,
    Polygon,
    box,
)
from shapely.validation import make_valid
from spherical_geometry.polygon import SphericalPolygon
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map

from srai.constants import GEOMETRY_COLUMN, WGS84_CRS

SPHERE_PARTS: list[SphericalPolygon] = []
SPHERE_PARTS_BOUNDING_BOXES: list[Polygon] = []

SCIPY_THRESHOLD = 1e-8


VertexHash = bytes
EdgeHash = tuple[VertexHash, VertexHash]


def _generate_sphere_parts() -> None:
    global SPHERE_PARTS, SPHERE_PARTS_BOUNDING_BOXES  # noqa: PLW0603

    if not SPHERE_PARTS:
        POINT_FRONT = (1.0, 0.0, 0.0)  # LON: 0; LAT: 0
        POINT_BACK = (-1.0, 0.0, 0.0)  # LON: 180; LAT: 0
        POINT_TOP = (0.0, 0.0, 1.0)  # LON: 0; LAT: 90
        POINT_BOTTOM = (0.0, 0.0, -1.0)  # LON: 0; LAT: -90
        POINT_LEFT = (0.0, -1.0, 0.0)  # LON: -90; LAT: 0
        POINT_RIGHT = (0.0, 1.0, 0.0)  # LON: 90; LAT: 0

        SPHERE_PARTS = [
            SphericalPolygon([POINT_FRONT, POINT_TOP, POINT_BACK, POINT_RIGHT, POINT_FRONT]),
            SphericalPolygon([POINT_FRONT, POINT_RIGHT, POINT_BACK, POINT_BOTTOM, POINT_FRONT]),
            SphericalPolygon([POINT_FRONT, POINT_BOTTOM, POINT_BACK, POINT_LEFT, POINT_FRONT]),
            SphericalPolygon([POINT_FRONT, POINT_LEFT, POINT_BACK, POINT_TOP, POINT_FRONT]),
        ]
        SPHERE_PARTS_BOUNDING_BOXES = [
            box(minx=0, miny=0, maxx=180, maxy=90),
            box(minx=0, miny=-90, maxx=180, maxy=0),
            box(minx=-180, miny=-90, maxx=0, maxy=0),
            box(minx=-180, miny=0, maxx=0, maxy=90),
        ]


def generate_voronoi_regions(
    seeds: Union[gpd.GeoDataFrame, list[Point]],
    max_meters_between_points: int = 10_000,
    num_of_multiprocessing_workers: int = -1,
    multiprocessing_activation_threshold: Optional[int] = None,
) -> list[MultiPolygon]:
    """
    Generate Thessien polygons for a given list of seeds.

    Function generates Thessien polygons on a sphere using
    SphericalVoronoi algorithm from scipy library.

    Args:
        seeds (Union[gpd.GeoDataFrame, List[Point]]): Seeds used for generating regions.
            If list, the points are expected to be in WGS84 coordinates (lat, lon).
            Otherwise, a GeoDataFrame will be transformed into WGS84.
        max_meters_between_points (int, optional): Maximal distance in meters between two points
            in the resulting polygon. Higher number results lower resolution of a polygon.
            Defaults to 10_000 (10 kilometers).
        num_of_multiprocessing_workers (int, optional): Number of workers used for multiprocessing.
            Defaults to -1 which results in a total number of available cpu threads.
            `0` and `1` values disable multiprocessing.
            Similar to `n_jobs` parameter from `scikit-learn` library.
        multiprocessing_activation_threshold (int, optional): Number of seeds required to start
            processing on multiple processes. Activating multiprocessing for a small
            amount of points might not be feasible. Defaults to 100.

    Returns:
        List[MultiPolygon]: List of MultiPolygons cut into up to 4 polygons based
            on quadrants of a sphere.

    Raises:
        ValueError: If less than 4 seeds are provided.
        ValueError: If any seed is duplicated.
        ValueError: If any seed is outside WGS84 coordinates domain.
    """
    if isinstance(seeds, gpd.GeoDataFrame):
        seeds, region_ids = _parse_geodataframe_seeds(seeds)
    else:
        region_ids = list(range(len(seeds)))

    if len(seeds) < 4:
        raise ValueError("Minimum 4 seeds are required.")

    duplicated_seeds_ids = _get_duplicated_seeds_ids(seeds, region_ids)
    if duplicated_seeds_ids:
        raise ValueError(f"Duplicate seeds present: {duplicated_seeds_ids}.")

    if not _check_if_in_bounds(seeds):
        raise ValueError("Seeds outside Earth WGS84 bounding box.")

    num_of_multiprocessing_workers = _parse_num_of_multiprocessing_workers(
        num_of_multiprocessing_workers
    )
    multiprocessing_activation_threshold = _parse_multiprocessing_activation_threshold(
        multiprocessing_activation_threshold
    )

    unit_sphere_ellipsoid = Ellipsoid(
        semimajor_axis=1, semiminor_axis=1, name="Unit Sphere", model="Unit"
    )
    mapped_points = [_map_to_geocentric(pt.x, pt.y, unit_sphere_ellipsoid) for pt in seeds]
    sphere_points = np.array([[pt[0], pt[1], pt[2]] for pt in mapped_points])

    radius = 1
    center = np.array([0, 0, 0])
    sv = SphericalVoronoi(sphere_points, radius, center, threshold=SCIPY_THRESHOLD)
    sv.sort_vertices_of_regions()

    total_regions = len(sv.regions)
    region_ids = list(range(total_regions))

    activate_multiprocessing = (
        num_of_multiprocessing_workers > 1 and total_regions >= multiprocessing_activation_threshold
    )

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)

        # generate all spherical polygons

        generate_spherical_polygons_parts_func = partial(
            _generate_spherical_polygons_parts,
            sv=sv,
        )

        if activate_multiprocessing:
            spherical_polygons_parts = [  # noqa: FURB179
                polygon_part_tuple
                for polygon_part_tuples in process_map(
                    generate_spherical_polygons_parts_func,
                    region_ids,
                    desc="Generating spherical polygons",
                    max_workers=num_of_multiprocessing_workers,
                    chunksize=ceil(total_regions / (4 * num_of_multiprocessing_workers)),
                )
                for polygon_part_tuple in polygon_part_tuples
            ]
        else:
            spherical_polygons_parts = [
                polygon_part_tuple
                for region_id in tqdm(region_ids, desc="Generating spherical polygons")
                for polygon_part_tuple in generate_spherical_polygons_parts_func(
                    region_id=region_id
                )
            ]

        # identify all edges

        hashed_vertices: dict[VertexHash, npt.NDArray[np.float32]] = {}
        hashed_edges: set[EdgeHash] = set()

        regions_parts: dict[int, list[tuple[int, list[EdgeHash]]]] = {}

        for region_id, sphere_part_id, spherical_polygon_points in spherical_polygons_parts:
            if region_id not in regions_parts:
                regions_parts[region_id] = []

            n = len(spherical_polygon_points)
            polygon_vertices_hashes = []
            polygon_edges = []

            # Hash all vertices
            for i in range(n):
                start: npt.NDArray[np.float32] = spherical_polygon_points[i]
                start_hash = hashlib.sha256(start.data).digest()
                hashed_vertices[start_hash] = start
                polygon_vertices_hashes.append(start_hash)

            # Map all edges
            for i in range(n):
                start_hash = polygon_vertices_hashes[i]
                end_hash = polygon_vertices_hashes[(i + 1) % n]

                if start_hash == end_hash:
                    continue

                polygon_edges.append((start_hash, end_hash))

                if (start_hash, end_hash) not in hashed_edges and (
                    end_hash,
                    start_hash,
                ) not in hashed_edges:
                    hashed_edges.add(
                        (
                            start_hash,
                            end_hash,
                        )
                    )

            regions_parts[region_id].append((sphere_part_id, polygon_edges))

        # interpolate unique ones

        interpolated_edges: dict[EdgeHash, list[tuple[float, float]]]

        interpolate_polygon_edge_func = partial(
            _interpolate_polygon_edge,
            hashed_vertices=hashed_vertices,
            ell=unit_sphere_ellipsoid,
            max_meters_between_points=max_meters_between_points,
        )

        if activate_multiprocessing:
            interpolated_edges = {
                hashed_edge: interpolated_edge
                for hashed_edge, interpolated_edge in zip(
                    hashed_edges,
                    process_map(
                        interpolate_polygon_edge_func,
                        hashed_edges,
                        desc="Interpolating edges",
                        max_workers=num_of_multiprocessing_workers,
                        chunksize=ceil(len(hashed_edges) / (4 * num_of_multiprocessing_workers)),
                    ),
                )
            }
        else:
            interpolated_edges = {
                hashed_edge: interpolate_polygon_edge_func(hashed_edge)
                for hashed_edge in tqdm(hashed_edges, desc="Interpolating edges")
            }

        # use interpolated edges to map spherical polygons into regions

        generated_regions: list[MultiPolygon] = []
        _generate_sphere_parts()

        for region_id in tqdm(region_ids, desc="Generating polygons"):
            multi_polygon_parts: list[Polygon] = []

            for sphere_part_id, region_polygon_edges in regions_parts[region_id]:
                polygon_points: list[tuple[float, float]] = []

                for edge_start, edge_end in region_polygon_edges:
                    if (edge_start, edge_end) in interpolated_edges:
                        interpolated_edge = interpolated_edges[(edge_start, edge_end)]
                    else:
                        interpolated_edge = interpolated_edges[(edge_end, edge_start)][::-1]

                    interpolated_edge = _fix_edge(
                        interpolated_edge,
                        SPHERE_PARTS_BOUNDING_BOXES[sphere_part_id].bounds,
                        prev_lon=polygon_points[-1][0] if polygon_points else None,
                        prev_lat=polygon_points[-1][1] if polygon_points else None,
                    )

                    polygon_points.extend(interpolated_edge)

                polygon = Polygon(polygon_points)
                polygon = make_valid(polygon)
                polygon = polygon.intersection(SPHERE_PARTS_BOUNDING_BOXES[sphere_part_id])
                if isinstance(polygon, GeometryCollection):
                    for geometry in polygon.geoms:
                        if isinstance(geometry, (Polygon, MultiPolygon)):
                            polygon = geometry
                            break
                    else:
                        raise RuntimeError(
                            f"Intersection with a quadrant did not produce any Polygon. ({polygon})"
                        )

                if isinstance(polygon, Polygon):
                    multi_polygon_parts.append(polygon)
                elif isinstance(polygon, MultiPolygon):
                    multi_polygon_parts.extend(polygon.geoms)
                elif isinstance(polygon, (LineString, MultiLineString, Point)):
                    pass
                else:
                    raise RuntimeError(str(type(polygon)))

            multi_polygon = MultiPolygon(multi_polygon_parts)
            generated_regions.append(multi_polygon)

    return generated_regions


def _parse_num_of_multiprocessing_workers(num_of_multiprocessing_workers: int) -> int:
    if num_of_multiprocessing_workers == 0:
        num_of_multiprocessing_workers = 1
    elif num_of_multiprocessing_workers < 0:
        num_of_multiprocessing_workers = cpu_count()

    return num_of_multiprocessing_workers


def _parse_multiprocessing_activation_threshold(
    multiprocessing_activation_threshold: Optional[int],
) -> int:
    if not multiprocessing_activation_threshold:
        multiprocessing_activation_threshold = 100

    return multiprocessing_activation_threshold


def _parse_geodataframe_seeds(gdf: gpd.GeoDataFrame) -> tuple[list[Point], list[Hashable]]:
    """Transform GeoDataFrame into list of Points with index."""
    seeds_wgs84 = gdf.to_crs(crs=WGS84_CRS)
    region_ids: list[Hashable] = []
    seeds: list[Point] = []

    for index, row in seeds_wgs84.iterrows():
        candidate_point = row.geometry.centroid
        if not candidate_point.is_empty:
            region_ids.append(index)
            seeds.append(candidate_point)

    return seeds, region_ids


def _get_duplicated_seeds_ids(seeds: list[Point], region_ids: list[Hashable]) -> list[Hashable]:
    """Return all seeds ids that overlap with another using quick sjoin operation."""
    gdf = gpd.GeoDataFrame(data={GEOMETRY_COLUMN: seeds}, index=region_ids, crs=WGS84_CRS)
    duplicated_seeds = gdf.sjoin(gdf).index.value_counts().loc[lambda x: x > 1]
    duplicated_seeds_ids: list[Hashable] = duplicated_seeds.index.to_list()
    return duplicated_seeds_ids


def _check_if_in_bounds(seeds: list[Point]) -> bool:
    """Check if all seeds are within bounds."""
    wgs84_earth_bbox = (box(minx=-180, miny=-90, maxx=180, maxy=90),)
    return all(point.covered_by(wgs84_earth_bbox) for point in seeds)


def _map_to_geocentric(lon: float, lat: float, ell: Ellipsoid) -> tuple[float, float, float]:
    """
    Wrapper for a geodetic2ecef function from pymap3d library.

    Args:
        lon (float): longitude of a point in a wgs84 crs.
        lat (float): latitude of a point in a wgs84 crs.
        ell (Ellipsoid): an ellipsoid.

    Returns:
        Tuple[float, float, float]: (x, y, z) coordinates tuple.
    """
    x, y, z = geodetic2ecef(lat, lon, 0, ell=ell)
    return x, y, z


def _generate_spherical_polygons_parts(
    region_id: int,
    sv: SphericalVoronoi,
) -> list[tuple[int, int, npt.NDArray[np.float32]]]:
    """
    Generate spherical polygon intersections with sphere parts.

    Args:
        region_id (int): Index of region in SphericalVoronoi result.
        sv (SphericalVoronoi): SphericalVoronoi object.

    Returns:
        List[Tuple[int, int, npt.NDArray[np.float32]]]: List of intersections containing
            region index, an index of a sphere part and a list of vertices.
    """
    region = sv.regions[region_id]
    region_vertices = np.array([v for v in sv.vertices[region]])

    sph_pol: Optional[SphericalPolygon] = None

    sphere_intersection_parts = []
    _generate_sphere_parts()

    for sphere_part_index, sphere_part in enumerate(SPHERE_PARTS):
        # check if whole region is inside
        if all(sphere_part.contains_point(point) for point in region_vertices):
            sphere_intersection_parts.append((region_id, sphere_part_index, region_vertices))
            # skip checking other sphere parts
            break
        # check if partially inside
        elif any(sphere_part.contains_point(point) for point in region_vertices):
            # delay SphericalPolygon creation since it's an expensive operation
            if not sph_pol:
                sph_pol = SphericalPolygon(region_vertices)

            intersection = sph_pol.intersection(sphere_part)
            for points in intersection.points:
                sphere_intersection_parts.append((region_id, sphere_part_index, points))

    # second check for the corner case when the region is on the intersection of 3 or 4 sphere parts
    # and sphere part only intersects a region's arc without vertex in it
    if len(sphere_intersection_parts) in (2, 3):
        for sphere_part_index, sphere_part in enumerate(SPHERE_PARTS):
            if any(
                sphere_intersection_part_tuple[1] == sphere_part_index
                for sphere_intersection_part_tuple in sphere_intersection_parts
            ):
                continue

            intersection = cast(SphericalPolygon, sph_pol).intersection(sphere_part)
            for points in intersection.points:
                sphere_intersection_parts.append((region_id, sphere_part_index, points))

    return sphere_intersection_parts


def _interpolate_polygon_edge(
    hashed_edge: EdgeHash,
    hashed_vertices: dict[VertexHash, npt.NDArray[np.float32]],
    ell: Ellipsoid,
    max_meters_between_points: int,
) -> list[tuple[float, float]]:
    """
    Interpolates spherical polygon arc edge to the latitude and longitude.

    Args:
        hashed_edge (EdgeHash): Edge hash containing start and end point.
        hashed_vertices (Dict[VertexHash, npt.NDArray[np.float32]]): Dict of hashed vertices.
        ell (Ellipsoid): Ellipsoid object.
        max_meters_between_points (int): Maximal distance between points.

    Returns:
        List[Tuple[float, float]]: _description_
    """
    start_point_hash, end_point_hash = hashed_edge
    start_point = hashed_vertices[start_point_hash]
    end_point = hashed_vertices[end_point_hash]

    longitudes, latitudes = _map_from_geocentric(
        [start_point[0], end_point[0]],
        [start_point[1], end_point[1]],
        [start_point[2], end_point[2]],
        ell,
    )
    start_lon, start_lat = longitudes[0], latitudes[0]
    end_lon, end_lat = longitudes[1], latitudes[1]
    haversine_distance = haversine((start_lat, start_lon), (end_lat, end_lon), unit="m")
    steps = max(ceil(haversine_distance / max_meters_between_points), 2)
    t_vals = np.linspace(0, 1, steps)

    edge_points = _interpolate_edge(
        start_point=start_point, end_point=end_point, step_ticks=t_vals, ell=ell
    )

    return edge_points


def _interpolate_edge(
    start_point: tuple[float, float, float],
    end_point: tuple[float, float, float],
    step_ticks: list[float],
    ell: Ellipsoid,
) -> list[tuple[float, float]]:
    """
    Generates latitude and longitude positions for an arc on the sphere.

    Args:
        start_point (Tuple[float, float, float]): Start position on an unit sphere.
        end_point (Tuple[float, float, float]): End position on an unit sphere.
        step_ticks (List[float]): Number of steps between.
        ell (Ellipsoid): Ellipsoid object.

    Returns:
        List[Tuple[float, float]]: List of latitude and longitude coordinates of the edge.
    """
    slerped_points = geometric_slerp(start_point, end_point, step_ticks, tol=SCIPY_THRESHOLD)

    xs = [pt[0] for pt in slerped_points]
    ys = [pt[1] for pt in slerped_points]
    zs = [pt[2] for pt in slerped_points]

    longitudes, latitudes = _map_from_geocentric(xs, ys, zs, ell)

    # round imperfections
    longitudes = np.round(longitudes, 8)
    latitudes = np.round(latitudes, 8)

    return [(longitude, latitude) for longitude, latitude in zip(longitudes, latitudes)]


def _fix_edge(
    edge_points: list[tuple[float, float]],
    bbox_bounds: tuple[float, float, float, float],
    prev_lon: Optional[float] = None,
    prev_lat: Optional[float] = None,
) -> list[tuple[float, float]]:
    """
    Fixes points laying on the edge between sphere parts.

    Args:
        edge_points (List[Tuple[float, float]]): Edge points to fix.
        bbox_bounds (Tuple[float, float, float, float]): Boundary box to use in the fixing.
        prev_lon (float, optional): Previous longitude of the edge. Can be passed to keep
            the context of the previous edges. Defaults to None.
        prev_lat (float, optional): Previous latitude of the edge. Can be passed to keep
            the context of the previous edges. Defaults to None.

    Returns:
        List[Tuple[float, float]]: Fixed edge points.
    """
    fixed_edge_points: list[tuple[float, float]] = []

    if prev_lon is not None and prev_lat is not None:
        fixed_edge_points.append((prev_lon, prev_lat))

    for longitude, latitude in edge_points:
        lon, lat = _fix_lat_lon(longitude, latitude, bbox_bounds)
        if prev_lon is not None and prev_lat is not None and abs(prev_lon - lon) >= 90:
            sign = 1 if lat > 0 else -1
            max_lat = sign * max(abs(prev_lat), abs(lat))
            if fixed_edge_points[-1] != (prev_lon, max_lat):
                fixed_edge_points.append((prev_lon, max_lat))
            if fixed_edge_points[-1] != (lon, lat):
                fixed_edge_points.append((lon, max_lat))
        fixed_edge_points.append((lon, lat))
        prev_lon = lon
        prev_lat = lat

    return fixed_edge_points


def _map_from_geocentric(
    x: npt.NDArray[np.float32],
    y: npt.NDArray[np.float32],
    z: npt.NDArray[np.float32],
    ell: Ellipsoid,
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]:
    """
    Wrapper for a ecef2geodetic function from pymap3d library.

    Args:
        x (npt.NDArray[np.float32]): X cartesian coordinate.
        y (npt.NDArray[np.float32]): Y cartesian coordinate.
        z (npt.NDArray[np.float32]): Z cartesian coordinate.
        ell (Ellipsoid): an ellipsoid.

    Returns:
        Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]: longitude and latitude coordinates
            in a wgs84 crs.
    """
    lat, lon, _ = ecef2geodetic_vectorized(x, y, z, ell=ell)
    return lon, lat


# Copyright (c) 2014-2022 Michael Hirsch, Ph.D.
# Copyright (c) 2013, Felipe Geremia Nievinski
# Copyright (c) 2004-2007 Michael Kleder
#
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions
# and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of
# conditions and the following disclaimer in the documentation and/or other materials provided
# with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
def ecef2geodetic_vectorized(
    x: npt.NDArray[np.float32],
    y: npt.NDArray[np.float32],
    z: npt.NDArray[np.float32],
    ell: Ellipsoid,
    deg: bool = True,
) -> npt.NDArray[np.float32]:
    """
    Modified function for mapping ecdf to geodetic values from ellipsoid.

    This function is a modified copy of `ecdf2geodetic` function from `pymap3d` library
    and is redistributed under BSD-2 license.

    Args:
        x (npt.NDArray[np.float32]): X coordinates.
        y (npt.NDArray[np.float32]): Y coordinates.
        z (npt.NDArray[np.float32]): Z coordinates.
        ell (Ellipsoid): Ellipsoid object.
        deg (bool, optional): Flag whether to return values in degrees. Defaults to True.

    Returns:
        npt.NDArray[np.float32]: Parsed latitudes and longitudes
    """
    with suppress(NameError):
        x = np.asarray(x)
        y = np.asarray(y)
        z = np.asarray(z)

    r = np.sqrt(x**2 + y**2 + z**2)

    E = np.sqrt(ell.semimajor_axis**2 - ell.semiminor_axis**2)

    # eqn. 4a
    u = np.sqrt(0.5 * (r**2 - E**2) + 0.5 * np.hypot(r**2 - E**2, 2 * E * z))

    Q = np.hypot(x, y)

    huE = np.hypot(u, E)

    # eqn. 4b
    try:
        with warnings.catch_warnings(record=False):
            warnings.simplefilter("error")
            Beta = np.arctan(huE / u * z / np.hypot(x, y))
    except (ArithmeticError, RuntimeWarning):
        is_zero_dimensions = len(x.shape) == 0

        if is_zero_dimensions:
            if np.isclose(z, 0):
                Beta = 0
            elif z > 0:
                Beta = np.pi / 2
            else:
                Beta = -np.pi / 2
        else:
            _beta_arr = []

            for _x, _y, _z, _u, _huE in zip(x, y, z, u, huE):
                try:
                    with warnings.catch_warnings(record=False):
                        warnings.simplefilter("error")
                        _beta = np.arctan(_huE / _u * _z / np.hypot(_x, _y))
                except (ArithmeticError, RuntimeWarning):
                    if np.isclose(_z, 0):
                        _beta = 0
                    elif _z > 0:
                        _beta = np.pi / 2
                    else:
                        _beta = -np.pi / 2
                _beta_arr.append(_beta)

            Beta = np.asarray(_beta_arr)

    # eqn. 13
    dBeta = ((ell.semiminor_axis * u - ell.semimajor_axis * huE + E**2) * np.sin(Beta)) / (
        ell.semimajor_axis * huE * 1 / np.cos(Beta) - E**2 * np.cos(Beta)
    )

    Beta += dBeta

    # eqn. 4c
    lat = np.arctan(ell.semimajor_axis / ell.semiminor_axis * np.tan(Beta))

    with suppress(NameError):
        # patch latitude for float32 precision loss
        lim_pi2 = np.pi / 2 - np.finfo(dBeta.dtype).eps
        lat = np.where(Beta >= lim_pi2, np.pi / 2, lat)
        lat = np.where(Beta <= -lim_pi2, -np.pi / 2, lat)

    lon = np.arctan2(y, x)

    # eqn. 7
    cosBeta = np.cos(Beta)
    with suppress(NameError):
        # patch altitude for float32 precision loss
        cosBeta = np.where(Beta >= lim_pi2, 0, cosBeta)
        cosBeta = np.where(Beta <= -lim_pi2, 0, cosBeta)

    alt = np.hypot(z - ell.semiminor_axis * np.sin(Beta), Q - ell.semimajor_axis * cosBeta)

    # inside ellipsoid?
    inside = (
        x**2 / ell.semimajor_axis**2 + y**2 / ell.semimajor_axis**2 + z**2 / ell.semiminor_axis**2
        < 1
    )

    try:
        if inside.any():
            # avoid all false assignment bug
            alt[inside] = -alt[inside]
    except (TypeError, AttributeError):
        if inside:
            alt = -alt

    if deg:
        lat = np.degrees(lat)
        lon = np.degrees(lon)

    return lat, lon, alt


def _fix_lat_lon(
    lon: float,
    lat: float,
    bbox: tuple[float, float, float, float],
) -> tuple[float, float]:
    """
    Fix point signs and rounding.

    Checks if any point is on a boundary and flips its sign
    to ensure validity of a polygon.

    Args:
        lon (float): Longitude of a point.
        lat (float): Latitude of a point.
        bbox (Tuple[float, float, float, float]): Current sphere octant bounding box.

    Returns:
        Tuple[float, float]: Longitude and latitude of a point.
    """
    min_lon, min_lat, max_lon, max_lat = bbox

    # switch signs
    if lon and abs(lon) == abs(min_lon):
        lon = min_lon
    elif lon and abs(lon) == abs(max_lon):
        lon = max_lon
    if lat and abs(lat) == abs(min_lat):
        lat = min_lat
    elif lat and abs(lat) == abs(max_lat):
        lat = max_lat

    return lon, lat



"""GeoVexEmbedder tests."""

import os
from contextlib import nullcontext as does_not_raise
from pathlib import Path
from typing import Any, Union, cast

import geopandas as gpd
import pandas as pd
import pytest
import torch
from pandas.testing import assert_frame_equal
from pytorch_lightning import seed_everything

from srai.embedders.geovex.embedder import GeoVexEmbedder
from srai.embedders.geovex.model import GeoVexModel
from srai.exceptions import ModelNotFitException
from srai.loaders.osm_loaders.filters import HEX2VEC_FILTER
from srai.neighbourhoods import H3Neighbourhood
from tests.embedders.geovex.constants import EMBEDDING_SIZE, PREDEFINED_TEST_CASES, TRAINER_KWARGS


@pytest.mark.parametrize(  # type: ignore
    "target_features, conv_layer_size, expectation",
    [
        (None, 256, does_not_raise()),
        (["building", "amenity"], 32, does_not_raise()),
        (["building", "amenity"], 256, pytest.raises(ValueError)),
    ],
)
def test_checking_encoder_sizes(
    target_features: Union[str, list[str]], conv_layer_size, expectation: Any
) -> None:
    """Test that incorrect encoder sizes raise ValueError."""
    target_tags: dict[str, list[str]] = target_features or HEX2VEC_FILTER  # type: ignore
    target_features = [f"{t}_{st}" for t in target_tags for st in HEX2VEC_FILTER[t]]  # type: ignore

    with expectation:
        GeoVexEmbedder(
            target_features=target_features,
            convolutional_layer_size=conv_layer_size,
        )


def test_embedder_not_fitted() -> None:
    """Test that GeoVexEmbedder raises ModelNotFitException if not fitted."""
    embedder = GeoVexEmbedder(
        [f"{t}_{st}" for t in HEX2VEC_FILTER for st in HEX2VEC_FILTER[t]],  # type: ignore
    )
    with pytest.raises(ModelNotFitException):
        embedder.transform(gpd.GeoDataFrame(geometry=[]), gpd.GeoDataFrame(), gpd.GeoDataFrame())


def test_embedder() -> None:
    """Test GeoVexEmbedder on predefined test cases."""
    test_files_path = Path(__file__).parent / "test_files"
    for test_case in PREDEFINED_TEST_CASES:
        name = test_case["test_case_name"]
        seed = test_case["seed"]
        radius: int = test_case["model_radius"]  # type: ignore
        print(name, seed)

        expected = pd.read_parquet(test_files_path / f"{name}_result.parquet")
        regions_gdf = gpd.read_parquet(test_files_path / f"{name}_regions.parquet")
        features_gdf = gpd.read_parquet(test_files_path / f"{name}_features.parquet")
        joint_gdf = pd.read_parquet(test_files_path / f"{name}_joint.parquet")
        seed_everything(seed, workers=True)
        os.environ["PYTHONHASHSEED"] = str(seed)
        torch.use_deterministic_algorithms(True)

        neighbourhood = H3Neighbourhood(regions_gdf)
        target_features = [
            f"{st}_{t}"
            for st in test_case["tags"]  # type: ignore
            for t in test_case["tags"][st]  # type: ignore
        ]
        embedder = GeoVexEmbedder(
            target_features=target_features,
            batch_size=10,
            neighbourhood_radius=radius,
            embedding_size=EMBEDDING_SIZE,
            convolutional_layers=test_case["num_layers"],  # type: ignore
            convolutional_layer_size=test_case["convolutional_layer_size"],  # type: ignore
        )

        counts_df, _, _ = embedder._prepare_dataset(
            regions_gdf, features_gdf, joint_gdf, neighbourhood, embedder._batch_size, shuffle=True
        )

        embedder._prepare_model(counts_df, 0.001)

        for _, param in cast(GeoVexModel, embedder._model).named_parameters():
            param.data.fill_(0.01)

        result_df = embedder.fit_transform(
            regions_gdf,
            features_gdf,
            joint_gdf,
            neighbourhood,
            trainer_kwargs=TRAINER_KWARGS,
            learning_rate=0.001,
        )
        result_df.columns = result_df.columns.astype(str)
        print(result_df.head())
        print(expected.head())
        assert_frame_equal(result_df, expected, atol=1e-1)



"""
Geofabrik layers filter.

This module contains the grouped OSM tags filter that is defined by Geofabrik [1].
Based on the document version `0.7.12`.

Note: not all definitions from the document are implemented, such as boundaries or places.
Note: buildings definition has been altered to include only tags that are recognized by TagInfo.

References:
    1. https://www.geofabrik.de/data/geofabrik-osm-gis-standard-0.7.pdf
"""

from srai.loaders.osm_loaders.filters import GroupedOsmTagsFilter

GEOFABRIK_LAYERS: GroupedOsmTagsFilter = {
    "public": {
        "amenity": [
            "police",
            "fire_station",
            "post_box",
            "post_office",
            "telephone",
            "library",
            "townhall",
            "courthouse",
            "prison",
            "embassy",
            "community_centre",
            "nursing_home",
            "arts_centre",
            "grave_yard",
            "marketplace",
            "recycling",
            "public_building",
        ],
        "office": ["diplomatic"],
        "landuse": ["cemetery"],
    },
    "education": {
        "amenity": [
            "university",
            "school",
            "kindergarten",
            "college",
        ]
    },
    "health": {
        "amenity": [
            "pharmacy",
            "hospital",
            "clinic",
            "doctors",
            "dentist",
            "veterinary",
        ]
    },
    "leisure": {
        "amenity": [
            "theatre",
            "nightclub",
            "cinema",
            "swimming_pool",
            "theatre",
            "theatre",
        ],
        "leisure": [
            "park",
            "playground",
            "dog_park",
            "sports_centre",
            "swimming_pool",
            "water_park",
            "golf_course",
            "stadium",
            "ice_rink",
        ],
        "sport": [
            "swimming",
            "tennis",
        ],
    },
    "catering": {
        "amenity": [
            "restaurant",
            "fast_food",
            "cafe",
            "pub",
            "bar",
            "food_court",
            "biergarten",
        ]
    },
    "accommodation": {
        "tourism": [
            "hotel",
            "motel",
            "bed_and_breakfast",
            "guest_house",
            "hostel",
            "chalet",
            "camp_site",
            "alpine_hut",
            "caravan_site",
        ],
        "amenity": ["shelter"],
    },
    "shopping": {
        "shop": [
            "supermarket",
            "bakery",
            "kiosk",
            "mall",
            "department_store",
            "general",
            "convenience",
            "clothes",
            "florist",
            "chemist",
            "books",
            "butcher",
            "shoes",
            "alcohol",
            "beverages",
            "optician",
            "jewelry",
            "gift",
            "sports",
            "stationery",
            "outdoor",
            "mobile_phone",
            "toys",
            "newsagent",
            "greengrocer",
            "beauty",
            "video",
            "car",
            "bicycle",
            "doityourself",
            "hardware",
            "furniture",
            "computer",
            "garden_centre",
            "hairdresser",
            "car_repair",
            "travel_agency",
            "laundry",
            "dry_cleaning",
        ],
        "amenity": ["car_rental", "car_wash", "car_sharing", "bicycle_rental", "vending_machine"],
        "vending": [
            "cigarettes",
            "parking_tickets",
        ],
    },
    "money": {"amenity": ["bank", "atm"]},
    "tourism": {
        "tourism": [
            "information",
            "attraction",
            "museum",
            "artwork",
            "picnic_site",
            "viewpoint",
            "zoo",
            "theme_park",
        ],
        "historic": [
            "monument",
            "memorial",
            "castle",
            "ruins",
            "archaeological_site",
            "wayside_cross",
            "wayside_shrine",
            "battlefield",
            "fort",
        ],
    },
    "miscpoi": {
        "amenity": [
            "toilets",
            "bench",
            "drinking_water",
            "fountain",
            "hunting_stand",
            "waste_basket",
            "emergency_phone",
            "fire_hydrant",
        ],
        "man_made": [
            "surveillance",
            "tower",
            "water_tower",
            "windmill",
            "lighthouse",
            "wastewater_plant",
            "water_well",
            "watermill",
            "water_works",
        ],
        "emergency": ["phone", "fire_hydrant"],
        "highway": ["emergency_access_point"],
    },
    "pofw": {"amenity": ["place_of_worship"]},
    "natural": {
        "natural": [
            "spring",
            "glacier",
            "peak",
            "cliff",
            "volcano",
            "tree",
            "mine",
            "cave_entrance",
            "beach",
        ]
    },
    "traffic": {
        "highway": [
            "traffic_signals",
            "mini_roundabout",
            "stop",
            "crossing",
            "ford",
            "motorway_junction",
            "turning_circle",
            "speed_camera",
            "street_lamp",
        ],
        "railway": ["level_crossing"],
    },
    "fuel_parking": {"amenity": ["fuel", "parking", "bicycle_parking"], "highway": ["services"]},
    "water_traffic": {
        "leisure": [
            "slipway",
            "marina",
        ],
        "man_made": ["pier"],
        "waterway": [
            "dam",
            "waterfall",
            "lock_gate",
            "weir",
        ],
    },
    "transport": {
        "railway": ["station", "halt", "tram_stop"],
        "public_transport": ["stop_position"],
        "highway": ["bus_stop"],
        "amenity": ["bus_station", "taxi", "ferry_terminal"],
        "aerialway": ["station"],
    },
    "air_traffic": {
        "amenity": ["airport"],
        "aeroway": [
            "aerodrome",
            "airfield",
            "aeroway",
            "helipad",
            "apron",
        ],
        "military": ["airfield"],
    },
    "major_roads": {
        "highway": [
            "motorway",
            "trunk",
            "primary",
            "secondary",
            "tertiary",
        ]
    },
    "minor_roads": {
        "highway": [
            "unclassified",
            "residential",
            "living_street",
            "pedestrian",
            "busway",
        ]
    },
    "highway_links": {
        "highway": [
            "motorway_link",
            "trunk_link",
            "primary_link",
            "secondary_link",
            "tertiary_link",
        ]
    },
    "very_small_roads": {
        "highway": [
            "service",
            "track",
        ]
    },
    "paths_unsuitable_for_cars": {
        "highway": ["bridleway", "path", "cycleway", "footway", "steps"],
        "cycle": ["designated"],
        "horse": ["designated"],
        "foot": ["designated"],
    },
    "unkown_roads": {"highway": ["road"]},
    "railways": {
        "railway": [
            "rail",
            "light_rail",
            "subway",
            "tram",
            "monorail",
            "narrow_gauge",
            "miniature",
            "funicular",
            "rack",
        ],
        "aerialway": [
            "drag_lift",
            "chair_lift",
            "high_speed_chair_lift",
            "cable_car",
            "gondola",
            "goods",
            "platter",
            "t-bar",
            "j-bar",
            "magic_carpet",
            "zip_line",
            "rope_tow",
            "mixed_lift",
        ],
    },
    "waterways": {
        "waterway": [
            "river",
            "stream",
            "canal",
            "drain",
        ]
    },
    "buildings": {
        "building": [
            "allotment_house",
            "apartments",
            "barn",
            "boathouse",
            "bungalow",
            "bunker",
            "cabin",
            "carport",
            "chapel",
            "church",
            "civic",
            "collapsed",
            "college",
            "commercial",
            "construction",
            "cowshed",
            "damaged",
            "detached",
            "dormitory",
            "farm",
            "farm_auxiliary",
            "fire_station",
            "garage",
            "garages",
            "ger",
            "government",
            "grandstand",
            "greenhouse",
            "hangar",
            "hospital",
            "hotel",
            "house",
            "houseboat",
            "hut",
            "industrial",
            "kindergarten",
            "kiosk",
            "manufacture",
            "mosque",
            "office",
            "outbuilding",
            "parking",
            "pavilion",
            "public",
            "residential",
            "retail",
            "roof",
            "ruins",
            "school",
            "semi",
            "semidetached_house",
            "service",
            "shed",
            "silo",
            "slurry_tank",
            "sports_centre",
            "stable",
            "static_caravan",
            "stilt_house",
            "storage_tank",
            "sty",
            "supermarket",
            "temple",
            "terrace",
            "toilets",
            "train_station",
            "transformer_tower",
            "transportation",
            "trullo",
            "university",
            "warehouse",
            "yes",
        ]
    },
    "landuse": {
        "landuse": [
            "forest",
            "residential",
            "industrial",
            "cemetery",
            "allotments",
            "meadow",
            "commercial",
            "recreation_ground",
            "retail",
            "military",
            "quarry",
            "orchard",
            "vineyard",
            "scrub",
            "grass",
            "military",
            "farmland",
            "farmyard",
        ],
        "leisure": ["park", "common", "nature_reserve", "recreation_ground"],
        "natural": ["wood", "heath"],
        "boundary": ["national_park"],
    },
    "water": {
        "natural": ["water", "glacier", "wetland"],
        "landuse": ["reservoir"],
        "waterway": ["riverbank", "dock"],
    },
}



"""
S2 Regionalizer.

This module exposes Google's S2 Geospatial Indexing System [1] as a regionalizer.
It uses the Python API [2].

References:
    1. https://s2geometry.io/
    2. https://github.com/JoaoCarabetta/s2-py
"""

import json
from typing import Any

import geopandas as gpd
from functional import seq
from functional.pipeline import Sequence
from s2 import s2
from shapely.geometry import Polygon

from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS
from srai.regionalizers import Regionalizer


class S2Regionalizer(Regionalizer):
    """
    S2 Regionalizer.

    S2 Regionalizer gives an opportunity to divide the given geometries into square S2 cells.
    """

    def __init__(self, resolution: int, buffer: bool = True) -> None:
        """
        Init S2 Regionalizer.

        Args:
            resolution (int): Resolution of the cells (S2 supports 0-30). See [1] for
                a full comparison.
            buffer (bool, optional): If True then fully cover geometries with S2 cells.
                Otherwise only use those cells that fully fit into them. Defaults to True.

        Raises:
            ValueError: If resolution is not between 0 and 30.

        References:
            1. https://s2geometry.io/resources/s2cell_statistics.html
        """
        if not (0 <= resolution <= 30):
            raise ValueError(f"Resolution {resolution} is not between 0 and 30.")

        self.resolution = resolution
        self.buffer = buffer

    def transform(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """
        Regionalize a given GeoDataFrame.

        Args:
            gdf (gpd.GeoDataFrame): GeoDataFrame to be regionalized.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with regionalized geometries.
        """
        gdf_wgs84 = gdf.to_crs(crs=WGS84_CRS)

        s2_gdf = self._fill_with_s2_cells(self._explode_multipolygons(gdf_wgs84))

        # s2 library fills also holes in Polygons, so here we remove redundant cells
        res: gpd.GeoDataFrame = gpd.sjoin(
            s2_gdf,
            gdf_wgs84,
            how="inner",
            predicate="intersects" if self.buffer else "within",
        ).drop(columns=["index_right"])

        res = res[~res.index.duplicated(keep="first")]

        res.index.name = REGIONS_INDEX

        return res

    def _fill_with_s2_cells(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        geo_json = json.loads(gdf.to_json())
        cells = (
            seq(geo_json["features"])
            .flat_map(lambda f: self._geojson_to_cells(f[GEOMETRY_COLUMN], self.resolution))
            .to_dict()
        )
        cells_gdf = gpd.GeoDataFrame(
            None,
            index=cells.keys(),
            geometry=list(cells.values()),
            crs=WGS84_CRS,
        )

        return cells_gdf

    def _geojson_to_cells(self, geo_json: dict[str, Any], res: int) -> Sequence:
        raw_cells = s2.polyfill(geo_json, res, with_id=True, geo_json_conformant=True)
        cells: Sequence = seq(raw_cells).map(lambda c: (c["id"], Polygon(c[GEOMETRY_COLUMN])))

        return cells



"""Conftest for OSM loaders."""

import geopandas as gpd
import pandas as pd
import pytest
from shapely.geometry import Point, Polygon

from srai.constants import FEATURES_INDEX, GEOMETRY_COLUMN, WGS84_CRS


@pytest.fixture  # type: ignore
def empty_area_gdf() -> gpd.GeoDataFrame:
    """Get a gdf with no geometry."""
    return gpd.GeoDataFrame(crs=WGS84_CRS, geometry=[])


@pytest.fixture  # type: ignore
def area_with_no_objects_gdf() -> gpd.GeoDataFrame:
    """Get a gdf that contains no OSM objects."""
    return gpd.GeoDataFrame(crs=WGS84_CRS, geometry=[Polygon([(3, 5), (3, 10), (7, 10), (7, 5)])])


@pytest.fixture  # type: ignore
def empty_result_gdf() -> gpd.GeoDataFrame:
    """Get empty OSMOnlineLoader result gdf."""
    result_index = pd.Index(data=[], name=FEATURES_INDEX, dtype="object")
    return gpd.GeoDataFrame(index=result_index, crs=WGS84_CRS, geometry=[])


@pytest.fixture  # type: ignore
def single_polygon_area_gdf() -> gpd.GeoDataFrame:
    """Get an example area gdf with with one polygon."""
    polygon_1 = Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])
    gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [polygon_1]}, crs=WGS84_CRS)
    return gdf


@pytest.fixture  # type: ignore
def two_polygons_area_gdf() -> gpd.GeoDataFrame:
    """Get an example area gdf with with two polygons."""
    polygon_1 = Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])
    polygon_2 = Polygon([(1, 1), (2, 2), (2, 1), (1, 0)])
    gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [polygon_1, polygon_2]}, crs=WGS84_CRS)
    return gdf


@pytest.fixture  # type: ignore
def amenities_gdf() -> gpd.GeoDataFrame:
    """Get an example gdf representing OSM objects with amenity tag."""
    return gpd.GeoDataFrame(
        {"amenity": ["restaurant", "restaurant", "bar"], "address": ["a", "b", "c"]},
        geometry=[Point(0, 0), Point(1, 1), Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])],
        index=pd.MultiIndex.from_arrays(
            arrays=[
                ["node", "node", "way"],
                [1, 2, 3],
            ],
            names=["element_type", "osmid"],
        ),
        crs=WGS84_CRS,
    )


@pytest.fixture  # type: ignore
def building_gdf() -> gpd.GeoDataFrame:
    """Get an example gdf representing OSM objects with building tag."""
    return gpd.GeoDataFrame(
        {"building": ["commercial", "retail"], "address": ["a", "c"]},
        geometry=[Point(0, 0), Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])],
        index=pd.MultiIndex.from_arrays(
            arrays=[
                ["node", "way"],
                [1, 3],
            ],
            names=["element_type", "osmid"],
        ),
        crs=WGS84_CRS,
    )


@pytest.fixture  # type: ignore
def expected_result_single_polygon() -> gpd.GeoDataFrame:
    """Get the expected result of a query with single polygon for testing."""
    return gpd.GeoDataFrame(
        {
            "amenity": ["restaurant"],
        },
        geometry=[Point(0, 0)],
        index=pd.Index(
            data=[
                "node/1",
            ],
            name=FEATURES_INDEX,
            dtype="object",
        ),
        crs=WGS84_CRS,
    )


@pytest.fixture  # type: ignore
def expected_result_gdf_simple() -> gpd.GeoDataFrame:
    """Get the expected result of a simple query for testing."""
    return gpd.GeoDataFrame(
        {
            "amenity": ["restaurant", "restaurant"],
        },
        geometry=[Point(0, 0), Point(1, 1)],
        index=pd.Index(
            data=[
                "node/1",
                "node/2",
            ],
            name=FEATURES_INDEX,
            dtype="object",
        ),
        crs=WGS84_CRS,
    )


@pytest.fixture  # type: ignore
def expected_result_gdf_complex() -> gpd.GeoDataFrame:
    """Get the expected result of complex query for testing."""
    return gpd.GeoDataFrame(
        {
            "amenity": ["restaurant", "restaurant", "bar"],
            "building": ["commercial", None, "retail"],
        },
        geometry=[Point(0, 0), Point(1, 1), Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])],
        index=pd.Index(
            data=[
                "node/1",
                "node/2",
                "way/3",
            ],
            name=FEATURES_INDEX,
            dtype="object",
        ),
        crs=WGS84_CRS,
    )
